#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import json
from argparse import ArgumentParser

import pydicom

from ..metadata import add_patient_age, dicom_to_json, drop_empty_tags, drop_fields_by_length


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("file", help="DICOM file to print")
    parser.add_argument("--output", "-o", help="output format", choices=["txt", "json"], default="txt")
    parser.add_argument(
        "--max_length", "-l", help="drop fields with values longer than MAX_LENGTH", default=100, type=int
    )
    parser.add_argument("--tags", "-t", nargs="+", help="specific tags", default=None)
    return parser


def main(args: argparse.Namespace) -> None:
    with pydicom.dcmread(args.file, specific_tags=args.tags, stop_before_pixels=True) as dcm:
        dcm = drop_empty_tags(dcm)
        add_patient_age(dcm)
        drop_fields_by_length(dcm, args.max_length, inplace=True)

        if args.output == "txt":
            print(f"Metadata for {args.file}")
            if args.tags:
                for v in dcm.values():
                    print(v)
            else:
                print(dcm)
        elif args.output == "json":
            print(json.dumps(dicom_to_json(dcm), indent=2))


def entrypoint():
    parser = get_parser()
    try:
        main(parser.parse_args())
    except AttributeError as e:
        if "DirectoryRecordSequence" in str(e):
            return
        raise e


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
from argparse import ArgumentParser
from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from typing import List, Optional, Tuple

import pydicom

from ..dicom import has_dicm_prefix
from ..tags import Tag
from ..types import ImageType, MammogramType, get_value


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path", default="./", help="path to find from")
    parser.add_argument("--name", "-n", default="*", help="glob pattern for filename")
    parser.add_argument("--jobs", "-j", default=4, help="parallelism")
    parser.add_argument(
        "--unique", "-u", default=False, action="store_true", help="only print the first match for a unique image type"
    )
    parser.add_argument("--include-numeric", default=False, action="store_true", help="include numeric fields")
    # TODO add a field to only return DICOMs with readable image data
    return parser


def check_file(path: Path, args: argparse.Namespace) -> Optional[Tuple[Path, int, str, str]]:
    if not path.is_file() or not has_dicm_prefix(path):
        return
    num_frames = 1
    try:
        tags = {Tag.ImageType, *MammogramType.get_required_tags()}
        with pydicom.dcmread(path, specific_tags=list(tags), stop_before_pixels=True) as dcm:
            img_type = ImageType.from_dicom(dcm)
            mammogram_type = MammogramType.from_dicom(dcm) if dcm.Modality == "MG" else None
            num_frames = get_value(dcm, Tag.NumberOfFrames, 1)
    except AttributeError:
        return

    final_image_type: List[Optional[str]] = [None]
    return path, num_frames, str(mammogram_type), img_type.simple_repr()


def main(args: argparse.Namespace) -> None:
    path = Path(args.path)
    if not path.is_dir():
        raise NotADirectoryError(path)

    seen_types = set()

    def callback(x: Future):
        result = x.result()
        if result is None:
            return
        path, num_frames, simple_type, image_type = result

        if image_type not in seen_types:
            print(f"{path} - {num_frames} - {simple_type} - {image_type}", flush=True)
            seen_types.add(image_type)

    futures: List[Future] = []
    with ThreadPoolExecutor(args.jobs) as tp:
        for match in path.rglob(args.name):
            f = tp.submit(check_file, match, args)
            f.add_done_callback(callback)
            futures.append(f)


def entrypoint():
    parser = get_parser()
    main(parser.parse_args())


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .collection import FILTER_REGISTRY, RecordCollection, RecordCreator, RecordFilter, record_iterator
from .helpers import SOPUID, ImageUID, SeriesUID, StudyUID, TransferSyntaxUID
from .protocols import (
    SupportsDataSetInfo,
    SupportsGenerated,
    SupportsManufacturer,
    SupportsPatientID,
    SupportsSite,
    SupportsStudyDate,
    SupportsStudyID,
    SupportsUID,
)
from .record import (
    HELPER_REGISTRY,
    RECORD_REGISTRY,
    DicomFileRecord,
    DicomImageFileRecord,
    FileRecord,
    MammogramFileRecord,
    RecordHelper,
    StandardizedFilename,
)


__all__ = [
    "SeriesUID",
    "StudyUID",
    "record_iterator",
    "FileRecord",
    "RecordCollection",
    "TransferSyntaxUID",
    "SOPUID",
    "ImageUID",
    "DicomFileRecord",
    "DicomImageFileRecord",
    "MammogramFileRecord",
    "RecordCreator",
    "RECORD_REGISTRY",
    "HELPER_REGISTRY",
    "RecordHelper",
    "FILTER_REGISTRY",
    "RecordFilter",
    "StandardizedFilename",
    "SupportsGenerated",
    "SupportsManufacturer",
    "SupportsPatientID",
    "SupportsStudyDate",
    "SupportsStudyID",
    "SupportsUID",
    "SupportsStudyID",
    "SupportsSite",
    "SupportsDataSetInfo",
]



from pathlib import Path
from typing import Final

import pydicom
import pytest

from dicom_utils.private import (
    MEDCOG_ADDR,
    MEDCOG_NAME,
    PRIVATE_ELEMENTS_DESCRIPTION,
    get_medcog_elements,
    get_year,
    hash_any,
    store_medcog_elements,
)


num_dicom_test_files: Final[int] = 3


@pytest.fixture(params=pydicom.data.get_testdata_files("*rgb*.dcm")[:num_dicom_test_files])  # type: ignore
def dicom_test_file(request) -> Path:
    return Path(request.param)


@pytest.mark.parametrize(
    "test_data",
    [
        (),
        {},
        [],
        "",
        0,
        None,
        1.0,
    ],
)
def test_hash_any(test_data) -> None:
    hash_any(test_data)


def test_store_value_hashes(dicom_test_file) -> None:
    ds = pydicom.dcmread(dicom_test_file)
    medcog_elements = get_medcog_elements(ds)
    store_medcog_elements(ds, medcog_elements)
    block = ds.private_block(MEDCOG_ADDR, MEDCOG_NAME)
    assert block[0].value == PRIVATE_ELEMENTS_DESCRIPTION
    for i, element in enumerate(medcog_elements):
        assert block[i + 1].VR == element.VR
        assert block[i + 1].value == element.value


@pytest.mark.parametrize(
    "date_string, expected_year",
    [
        ("20200101", "2020"),
        ("19901231", "1990"),
        ("", "????"),
    ],
)
def test_get_year(test_dicom, date_string, expected_year) -> None:
    test_dicom.StudyDate = date_string
    assert expected_year == get_year(test_dicom).value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import runpy
import shutil
import sys
from pathlib import Path
from typing import Any

import numpy as np
import pydicom
import pytest
from pydicom.data import get_testdata_file


@pytest.fixture(
    params=[
        "JPEG2000.dcm",
        "CT_small.dcm",
        "JPEG-LL.dcm",
    ]
)
def dicom_filepath(request: Any) -> str:
    filepath = get_testdata_file(request.param)
    assert isinstance(filepath, str), filepath
    return filepath


def test_merge_series(tmp_path: Path, dicom_filepath: str) -> None:
    num_instances = 2
    for i in range(num_instances):
        shutil.copy(dicom_filepath, tmp_path / f"{i}.dcm")

    sys.argv = [sys.argv[0], str(tmp_path)]
    runpy.run_module("dicom_utils.cli.merge_series", run_name="__main__", alter_sys=True)

    ds = pydicom.dcmread(dicom_filepath)
    ds_merged = pydicom.dcmread(tmp_path / "merged-0.dcm")

    chns, rows, cols = ds_merged.pixel_array.shape
    assert ds_merged.NumberOfFrames == num_instances == chns
    assert ds.pixel_array.shape == (rows, cols)

    expected_pixel_array = np.repeat(ds.pixel_array[np.newaxis, :, :], num_instances, axis=0)
    assert np.allclose(ds_merged.pixel_array, expected_pixel_array)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
from abc import ABC, abstractmethod
from copy import deepcopy
from itertools import islice
from typing import Any, Callable, Iterator, List, Optional, Protocol, Sized, SupportsInt, Tuple, TypeVar, Union, cast

import numpy as np
from numpy.typing import NDArray
from pydicom import Dataset
from pydicom.encaps import encapsulate, generate_frames
from pydicom.pixel_data_handlers import numpy_handler
from pydicom.pixels.utils import reshape_pixel_array
from pydicom.uid import ImplicitVRLittleEndian
from registry import Registry


class SupportsGetItem(Protocol):
    def __getitem__(self, key: Any) -> Any: ...


T = TypeVar("T", bound=SupportsGetItem | NDArray)
U = TypeVar("U", bound=Dataset)


def dicom_copy(dcm: U) -> U:
    # Avoid multiple copies of PixelData which can be 100s of MB
    pixel_data = dcm.PixelData
    del dcm.PixelData
    if hasattr(dcm, "_pixel_array"):
        del dcm._pixel_array  # Delete possibly cached interpretation of PixelData
    new_dcm = deepcopy(dcm)
    dcm.PixelData = new_dcm.PixelData = pixel_data
    return new_dcm


VOLUME_HANDLERS = Registry("volume handlers")


class VolumeHandler(ABC):
    r"""Base class for classes that manipulate 3D Volumes"""

    @abstractmethod
    def get_indices(self, total_frames: Optional[int]) -> Tuple[int, Optional[int], int]: ...

    def __call__(self, x: Union[T, U]) -> Union[T, U]:
        if isinstance(x, Dataset):
            return self.handle_dicom(x)
        else:
            return self.handle_array(x)

    def handle_dicom(self, dcm: U) -> U:
        num_frames: Optional[SupportsInt] = dcm.get("NumberOfFrames", None)
        num_frames = int(num_frames) if num_frames is not None else None
        start, stop, stride = self.get_indices(num_frames)
        return self.slice_dicom(dcm, start, stop, stride)

    def handle_array(self, x: T) -> T:
        num_frames = len(x) if isinstance(x, Sized) else None
        start, stop, stride = self.get_indices(num_frames)
        return cast(T, self.slice_array(cast(SupportsGetItem, x), start, stop, stride))

    @classmethod
    def iterate_frames(cls, dcm: Dataset) -> Iterator[bytes]:
        is_compressed: bool = dcm.file_meta.TransferSyntaxUID.is_compressed
        num_frames: Optional[SupportsInt] = dcm.get("NumberOfFrames", None)
        num_frames = int(num_frames) if num_frames is not None else None
        if is_compressed:
            for frame in generate_frames(dcm.PixelData, number_of_frames=num_frames):
                yield frame
        else:
            # manually call the numpy handler so we can read the full array as read only.
            # this avoid memory duplication
            arr = numpy_handler.get_pixeldata(dcm, read_only=True)
            arr = reshape_pixel_array(dcm, arr)
            for frame in arr:
                yield frame.tobytes()

    @classmethod
    def slice_array(cls, x: T, start: int, stop: Optional[int], stride: int) -> T:
        r"""Slices an array input according to :func:`get_indices`"""
        len(x) if isinstance(x, Sized) else None
        result = x[slice(start, stop, stride)]
        return cast(T, result)

    @classmethod
    def update_pixel_data(cls, dcm: U, frames: List[bytes], preserve_compression: bool = True) -> U:
        r"""Updates PixelData with a new sequence of frames, accounting for compression type"""
        is_compressed: bool = dcm.file_meta.TransferSyntaxUID.is_compressed
        if is_compressed and preserve_compression:
            new_pixel_data = encapsulate(frames)
        else:
            new_pixel_data = b"".join(frames)
            # if dcm was compressed, we cant guarantee that the libraries needed to compress new
            # frames to that transfer syntax will be available. to account for this, just change
            # the old TSUID to an uncompressed variant and attach frames without compression
            if is_compressed:
                dcm.file_meta.TransferSyntaxUID = ImplicitVRLittleEndian
        dcm.PixelData = new_pixel_data
        dcm.NumberOfFrames = len(frames)
        return dcm

    @classmethod
    def slice_dicom(cls, dcm: U, start: int, stop: Optional[int], stride: int) -> U:
        r"""Slices a DICOM object input according to :func:`get_indices`.

        .. note:
            Unlike :func:`slice_array`, this function can perform slicing on compressed DICOMs
            with out needing to decompress all frames. This can provide a substantial performance gain.

        """
        # copy dicom and read key tags
        dcm = dicom_copy(dcm)
        num_frames: Optional[SupportsInt] = dcm.get("NumberOfFrames", None)
        num_frames = int(num_frames) if num_frames is not None else None
        is_compressed: bool = dcm.file_meta.TransferSyntaxUID.is_compressed

        # read sliced frames
        frame_iterator = cls.iterate_frames(dcm)
        frames = list(islice(frame_iterator, start, stop, stride))
        if not frames:
            raise IndexError("No frames remain in the sliced DICOM")
        if not all(frames):
            raise IndexError("One or more frames had no contents")

        # update dicom object and return
        dcm = cls.update_pixel_data(dcm, frames)
        assert dcm.NumberOfFrames == len(frames)
        return dcm


class KeepVolume(VolumeHandler):
    r"""Retains the entire input volume"""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def get_indices(self, total_frames: Optional[int]) -> Tuple[int, Optional[int], int]:
        return 0, None, 1


class SliceAtLocation(VolumeHandler):
    r"""Samples the input volume at centered on a given slice with optional context frames.

    Args:
        center:
            The slice about which to sample. Defaults to num_frames / 2

        before:
            Optional frames to sample before ``center``.

        after:
            Optional frames to sample after ``center``.

        stride:
            If given, the stride between sampled frames
    """

    def __init__(
        self,
        center: Optional[int] = None,
        before: int = 0,
        after: int = 0,
        stride: int = 1,
    ):
        self.center = center
        self.stride = stride
        self.before = before
        self.after = after

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}("
        s += f"center={self.center}, "
        s += f"before={self.before}, "
        s += f"stride={self.stride}"
        s += ")"
        return s

    def handle_dicom(self, dcm: U) -> U:
        num_frames: Optional[SupportsInt] = dcm.get("NumberOfFrames", None)
        if num_frames is None and self.center is None:
            raise AttributeError(f"`NumberOfFrames` cannot be absent when `{self.__class__.__name__}.center` is `None`")
        return super().handle_dicom(dcm)

    def get_indices(self, total_frames: Optional[int]) -> Tuple[int, int, int]:
        if self.center is None and total_frames is None:
            raise ValueError(f"`total_frames` cannot be `None` when `{self.__class__.__name__}.center` is `None`")
        center = self.center if self.center is not None else cast(int, total_frames) // 2
        start = center - self.before
        end = center + self.after + 1
        return start, end, self.stride


class UniformSample(VolumeHandler):
    r"""Samples the input volume at centered on a given slice with optional context frames.
    Either ``stride`` or ``count`` must be provided.

    Args:
        amount:
            When when ``method='count'``, the number of frames to sample.
            When ``method='stride'``, the stride between sampled frames.

        method:
            Either ``'count'`` or ``'stride'``.
    """

    def __init__(
        self,
        amount: int,
        method: str = "count",
    ):
        if method not in ("count", "stride"):
            raise ValueError(f"`method` {method} must be one of 'count', 'stride'")
        self.amount = amount
        self.method = method

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}("
        s += f"amount={self.amount}, "
        s += f"method={self.method}"
        s += ")"
        return s

    def get_indices(self, total_frames: Optional[int]) -> Tuple[int, Optional[int], int]:
        if self.method == "stride":
            return 0, None, self.amount
        elif self.method == "count":
            assert total_frames is not None
            stride = max(total_frames // self.amount, 1)
            return 0, None, stride
        else:
            raise ValueError(f"Invalid method {self.method}")


class ReduceVolume(VolumeHandler):
    r"""Reduces the input volume to a subset of frames by applying a reduction function. This volume handler
    will need to decompress frames in order to apply the reduction function. To support this, NVJPEG parameters
    are accepted by :func:`__call__` and passed to :func:`dicom_utils.dicom.decompress`.

    Args:
        reduction:
            A function that takes two frames and an ``out`` argument and returns the reduced frame.
            The ``out`` argument is used to avoid allocating new memory for the reduced frame.

        output_frames:
            The number of frames to output. Defaults to ``1``, meaning the entire volume is reduced to a single frame.

        skip_edge_frames:
            The number of frames to skip at the beginning and end of the volume. This is useful for removing
            artifacts from the beginning and end of the volume.

    Raises:
        RuntimeError: If no frames remain in the sliced DICOM
    """

    def __init__(
        self,
        reduction: Callable[..., np.ndarray] = cast(Any, np.maximum),
        output_frames: int = 1,
        skip_edge_frames: int = 0,
    ):
        self.reduction = reduction
        self.output_frames = output_frames
        self.skip_edge_frames = skip_edge_frames

    def __call__(self, x: Union[T, U], use_nvjpeg: Optional[bool] = None, **kwargs) -> Union[T, U]:
        if isinstance(x, Dataset):
            return self.handle_dicom(x, use_nvjpeg=use_nvjpeg, **kwargs)
        else:
            return self.handle_array(x)

    def get_indices(self, total_frames: Optional[int]) -> Tuple[int, int, int]:
        raise NotImplementedError

    def handle_dicom(self, dcm: U, use_nvjpeg: Optional[bool] = None, **kwargs) -> U:
        chunks = [chunk.tobytes() for chunk in self.iterate_chunks(dcm, self.reduction, use_nvjpeg, **kwargs)]
        if not chunks:
            raise RuntimeError("No frames remain in the sliced DICOM")
        assert len(chunks) == self.output_frames

        dcm = self.update_pixel_data(dcm, chunks, preserve_compression=False)
        assert not dcm.file_meta.TransferSyntaxUID.is_compressed, "DICOM should be decompressed after ReduceVolume"
        return dcm

    def handle_array(self, x: T) -> T:
        raise NotImplementedError

    def iterate_chunks(
        self,
        dcm: Dataset,
        reduction: Optional[Callable[..., np.ndarray]],
        use_nvjpeg: Optional[bool] = None,
        **kwargs,
    ) -> Iterator[np.ndarray]:
        r"""Iterates through the DICOM volume, yielding chunks based on the reduction parameters.
        If a reduction is given, each chunk will be reduced before being yielded.
        """
        # TODO: This is a bit of a hack to avoid circular imports. Fixing it will require a larger refactor.
        from .dicom import decompress

        start = self.skip_edge_frames
        stop = dcm.NumberOfFrames - self.skip_edge_frames

        # If the chunk size is less than 1, we need to adjust the start and stop indices.
        # We do this by widening the range between start and stop until we have a chunk size of 1
        # or until we reach the end of the volume. Since we are widening symetrically, we clip
        # the start index to a minimum of 0 so every possible frame is tried.
        while (chunk_size := (stop - start) // self.output_frames) < 1:
            start = max(start - 1, 0)
            stop += 1
            if stop > dcm.NumberOfFrames:
                raise RuntimeError("Could not find a valid chunk size")
        if chunk_size < 1:
            # TODO: consider returning the original volume + duplicating frames to fill the output
            raise RuntimeError("Could not find a valid chunk size")

        for yield_count in range(self.output_frames):
            # If this is the last chunk, make sure it includes all remaining frames
            is_last_chunk = yield_count == self.output_frames - 1
            chunk_end = stop if is_last_chunk else start + chunk_size

            # Slice the chunk
            assert start < chunk_end <= stop, f"Invalid chunk start/end: {start}, {chunk_end}, {stop}"
            sliced = self.slice_dicom(dcm, start, chunk_end, stride=1)

            # If no reduction is given, just yield the decompressed chunk
            if reduction is None or chunk_size == 1:
                sliced = decompress(sliced, use_nvjpeg=use_nvjpeg, **kwargs)
                yield sliced.pixel_array

            # Iterate through the chunk, applying the reduction in-place.
            # We never want to store more of the volume than is needed to conserve memory.
            else:
                arr = self.slice_dicom(sliced, 0, 1, stride=1).pixel_array
                for i in range(1, sliced.NumberOfFrames):
                    # TODO: it is less computationally efficient to decompress each frame individually,
                    # but it is more memory efficient. Memory efficiency is more important given the limitations
                    # of the deployment environment. Consider a better solution.
                    other = self.slice_dicom(sliced, i, i + 1, 1)
                    other = decompress(other, use_nvjpeg=use_nvjpeg, **kwargs)
                    arr = reduction(arr, other.pixel_array, out=arr)
                yield arr

            start += chunk_size


class RandomSlice(VolumeHandler):
    r"""Samples the input volume at a random slice

    Args:
        seed: The random seed to use
    """

    def __init__(self, seed: int = 42):
        self.rng = random.Random(seed)

    def get_indices(self, total_frames: Optional[int]) -> Tuple[int, int, int]:
        if total_frames is None:
            raise ValueError("`total_frames` cannot be `None`")
        index = self.rng.randint(0, total_frames - 1)
        return index, index + 1, 1


# Register some default handlers
VOLUME_HANDLERS(name="keep")(KeepVolume)
VOLUME_HANDLERS(name="max")(ReduceVolume)
VOLUME_HANDLERS(name="mean", reduction=np.mean)(ReduceVolume)
VOLUME_HANDLERS(name="slice")(SliceAtLocation)

# Multi-frame reductions
for output_frames in (1, 8, 10, 16):
    for skip_edge_frames in (0, 5, 10):
        VOLUME_HANDLERS(
            name=f"max-{output_frames}-{skip_edge_frames}",
            output_frames=output_frames,
            skip_edge_frames=skip_edge_frames,
        )(ReduceVolume)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import importlib as imp
import os
import warnings

import pytest

import dicom_utils
from dicom_utils import SILENCE_ENV_VAR, SPAM_WARNING_PATTERNS, filter_spam_warnings


@pytest.mark.parametrize("pattern", SPAM_WARNING_PATTERNS)
def test_filtered_patterns(mocker, pattern):
    # NOTE: pytest.warns seems to ignore filters, so test using mock calls instead
    spy = mocker.spy(warnings, "filterwarnings")
    filter_spam_warnings()
    return any(call.args == ("ignore", pattern) for call in spy.mock_calls)


@pytest.mark.parametrize("value", ["", "0", "1"])
def test_ignore_spam_env_var(mocker, value):
    # NOTE: mock on filter_spam_warnings() would be reset during import, so use filterwarnings
    os.environ[SILENCE_ENV_VAR] = value
    spy = mocker.spy(warnings, "filterwarnings")
    imp.reload(dicom_utils)
    ignored_all_spam = all(any(call.args == ("ignore", p) for call in spy.mock_calls) for p in SPAM_WARNING_PATTERNS)
    should_ignore_all_spam = value == "1"
    assert ignored_all_spam == should_ignore_all_spam



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path

import numpy as np
import pydicom
import pytest
from pydicom import Dataset, Sequence
from pydicom.uid import ExplicitVRLittleEndian, ImplicitVRLittleEndian, RLELossless
from pydicom.valuerep import VR

from dicom_utils import DicomFactory
from dicom_utils.dicom import set_pixels
from dicom_utils.dicom_factory import FACTORY_REGISTRY, CompleteMammographyStudyFactory
from dicom_utils.tags import Tag
from dicom_utils.types import MammogramType


class TestDicomFactory:
    def test_init(self):
        DicomFactory()

    def test_constructor_overrides(self):
        factory = DicomFactory(Modality="MG")
        dcm = factory()
        assert dcm.Modality == "MG"

    def test_override_present_value(self):
        factory = DicomFactory()
        dcm = factory(Modality="MG")
        assert dcm.Modality == "MG"

    def test_set_new_value(self):
        factory = DicomFactory()
        dcm = factory(ViewPosition="MLO")
        assert dcm.ViewPosition == "MLO"

    def test_set_code_sequence(self):
        factory = DicomFactory()
        vcs = DicomFactory.code_sequence("view-code")
        vcms = DicomFactory.code_sequence("view-modifier-code")
        dcm = factory(ViewCodeSequence=vcs, ViewModifierCodeSequence=vcms)
        assert dcm.ViewCodeSequence == vcs
        assert dcm.ViewModifierCodeSequence == vcms

    def test_pixel_array_from_dicom(self):
        factory = DicomFactory()
        arr = DicomFactory.pixel_array_from_dicom(factory.dicom)
        assert isinstance(arr, np.ndarray)
        assert arr.shape == (128, 128)
        # Should be deterministic for a given seed
        assert arr.min() == 9
        assert arr.max() == 65523

    @pytest.mark.parametrize("meaning", ["foo", "bar"])
    def test_code_sequence(self, meaning):
        seq = DicomFactory.code_sequence(meaning)
        assert isinstance(seq, Sequence)
        assert len(seq) == 1
        assert seq[0].CodeMeaning == meaning

    @pytest.mark.parametrize(
        "tag,value,exp",
        [
            pytest.param(Tag.PatientAge, "18Y", VR("AS")),
            pytest.param(Tag.StudyDate, "00010101", VR("DA")),
            pytest.param(Tag.StudyTime, "010101", VR("TM")),
            pytest.param(Tag.StudyInstanceUID, "1.2.345", VR("UI")),
            pytest.param(Tag.PhotometricInterpretation, "MONOCHROME1", VR("CS")),
            pytest.param(Tag.ViewCodeSequence, [Dataset()], VR("SQ")),
        ],
    )
    def test_suggest_vr(self, tag, value, exp):
        vr = DicomFactory.suggest_vr(tag, value)
        assert vr == exp

    def test_ffdm_factory(self):
        factory = FACTORY_REGISTRY.get("ffdm")()
        dcm = factory()
        assert MammogramType.from_dicom(dcm) == MammogramType.FFDM
        assert dcm.pixel_array.any()

    def test_tomo_factory(self):
        factory = FACTORY_REGISTRY.get("tomo")()
        dcm = factory()
        assert MammogramType.from_dicom(dcm) == MammogramType.TOMO
        assert dcm.pixel_array.any()

    def test_synth_factory(self):
        factory = FACTORY_REGISTRY.get("synth")()
        dcm = factory()
        assert MammogramType.from_dicom(dcm) == MammogramType.SYNTH
        assert dcm.pixel_array.any()

    def test_ultrasound_factory(self):
        factory = FACTORY_REGISTRY.get("ultrasound")()
        dcm = factory()
        assert dcm.Modality == "US"
        assert dcm.pixel_array.any()

    def test_complete_mammography_case_factory(self):
        factory = FACTORY_REGISTRY.get("mammo-case")()
        dicoms = factory()
        assert len(dicoms) == 12
        for dcm in dicoms:
            assert dcm.pixel_array.any()

    @pytest.mark.parametrize("allow", [True, False])
    def test_nonproto_tag(self, allow):
        factory = DicomFactory(allow_nonproto_tags=allow)
        assert Tag.SeriesDescription not in factory.dicom
        dcm = factory(SeriesDescription="foo")

        if allow:
            assert dcm.SeriesDescription == "foo"
        else:
            assert Tag.SeriesDescription not in dcm

    def test_skip_pixels(self):
        factory = FACTORY_REGISTRY.get("ffdm")()
        dcm = factory(pixels=False)
        assert not hasattr(dcm, "PixelData")


class TestCompleteMammographyStudyFactory:

    @pytest.fixture
    def dicoms(self):
        H, W = 64, 32
        fact = CompleteMammographyStudyFactory(
            Rows=H,
            Columns=W,
            seed=42,
            BitsAllocated=16,
            WindowCenter=4096,
            WindowWidth=8192,
        )
        dicoms = fact()

        # Ensure we have a mix of inversions
        dicoms[0].PhotometricInterpretation = "MONOCHROME1"
        dicoms[1].PhotometricInterpretation = "MONOCHROME2"

        # Ensure we have a mix of TSUIDs
        tsuids = (ImplicitVRLittleEndian, ExplicitVRLittleEndian, RLELossless)
        for dcm, tsuid in zip(dicoms[: len(tsuids)], tsuids):
            set_pixels(dcm, dcm.pixel_array, tsuid)
            dcm.file_meta.TransferSyntaxUID = tsuid

        return dicoms

    def test_dicoms(self, dicoms):
        assert len(dicoms) == 12
        for dcm in dicoms:
            assert dcm.Rows == 64
            assert dcm.Columns == 32
            assert dcm.BitsAllocated == 16
            assert dcm.WindowCenter == 4096
            assert dcm.WindowWidth == 8192
            assert dcm.NumberOfFrames in (1, 3)
            assert dcm.pixel_array is not None

    def test_save_load_dicoms(self, tmp_path, dicoms):
        for dcm in dicoms:
            dcm: pydicom.FileDataset
            dest = Path(tmp_path, f"{dcm.SOPInstanceUID}.dcm")
            dcm.save_as(dest)

        loaded = list(Path(tmp_path).glob("*.dcm"))
        assert len(loaded) == 12
        for path in loaded:
            dcm = pydicom.dcmread(path)
            assert dcm.Rows == 64
            assert dcm.Columns == 32
            assert dcm.BitsAllocated == 16
            assert dcm.WindowCenter == 4096
            assert dcm.WindowWidth == 8192
            assert dcm.NumberOfFrames in (1, 3)
            assert dcm.pixel_array is not None
            if dcm.NumberOfFrames > 1:
                assert dcm.pixel_array.shape[0] == dcm.NumberOfFrames



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import runpy
import shutil
import sys
from pathlib import Path
from typing import List

import numpy as np
import pydicom
import pytest
from PIL import Image
from pydicom.dataset import Dataset, FileMetaDataset
from pydicom.sequence import Sequence


@pytest.fixture
def dicom_image_file(tmpdir) -> str:
    filename = pydicom.data.get_testdata_file("CT_small.dcm")  # type: ignore
    shutil.copy(filename, tmpdir)
    return filename


def create_referenced_image_sequence(ref_ds: Dataset) -> Sequence:
    refd_image = Dataset()
    refd_image.ReferencedSOPClassUID = ref_ds.SOPClassUID
    refd_image.ReferencedSOPInstanceUID = ref_ds.SOPInstanceUID
    refd_image.ReferencedFrameNumber = "1"
    return Sequence([refd_image])


def create_graphic_object_sequence(data: List[float] = [0, 0, 1, 1], graphic_type: str = "CIRCLE") -> Sequence:
    graphic_object = Dataset()
    graphic_object.GraphicAnnotationUnits = "PIXEL"
    graphic_object.GraphicDimensions = 2
    graphic_object.GraphicData = data
    graphic_object.GraphicType = graphic_type
    return Sequence([graphic_object])


def create_graphic_annotation_sequence(ref_ds: Dataset) -> Sequence:
    graphic_annotation = Dataset()
    graphic_annotation.ReferencedImageSequence = create_referenced_image_sequence(ref_ds)
    graphic_annotation.GraphicObjectSequence = create_graphic_object_sequence()
    return Sequence([graphic_annotation])


@pytest.fixture
def dicom_annotation_file(tmpdir, dicom_image_file) -> None:
    ds = Dataset()
    ds.Modality = "PR"
    ds.GraphicAnnotationSequence = create_graphic_annotation_sequence(pydicom.dcmread(dicom_image_file))
    ds.file_meta = FileMetaDataset()
    ds.file_meta.TransferSyntaxUID = pydicom.uid.ExplicitVRLittleEndian  # type: ignore
    ds.save_as(tmpdir / "annotation.dcm", enforce_file_format=False)


@pytest.mark.parametrize("out", [None, "foo.png"])
def test_dicom2img(dicom_image_file, dicom_annotation_file, tmp_path, out):
    sys.argv = [sys.argv[0], str(tmp_path), "--noblock"]

    if out is not None:
        path = Path(tmp_path, out)
        sys.argv.extend(["--output", str(path)])

    runpy.run_module("dicom_utils.cli.dicom2img", run_name="__main__", alter_sys=True)

    if out is not None:
        assert "path" in locals()
        assert locals()["path"].is_file()


@pytest.mark.parametrize(
    "out", [pytest.param(None, marks=pytest.mark.xfail(raises=NotImplementedError, strict=True)), "foo.gif"]
)
@pytest.mark.parametrize("handler", ["keep", "max-8-5"])
def test_dicom2img_3d(dicom_file_3d, tmp_path, out, handler):
    sys.argv = [sys.argv[0], str(tmp_path), "--noblock", "-v", handler]

    if out is not None:
        path = Path(tmp_path, out)
        sys.argv.extend(["--output", str(path)])

    runpy.run_module("dicom_utils.cli.dicom2img", run_name="__main__", alter_sys=True)

    if out is not None:
        assert "path" in locals()
        assert locals()["path"].is_file()


def test_bytes(dicom_image_file, dicom_annotation_file, tmp_path, capsysbinary):
    sys.argv = [sys.argv[0], str(tmp_path), "--noblock", "--bytes"]

    runpy.run_module("dicom_utils.cli.dicom2img", run_name="__main__", alter_sys=False)
    captured = capsysbinary.readouterr().out
    dest = Path(tmp_path, "out.png")
    with open(dest, "wb") as f:
        f.write(captured)

    img = np.asarray(Image.open(dest))
    assert img.shape == (128, 128, 3)



import copy
import logging
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List

import pydicom
from pydicom import Dataset
from pydicom.errors import InvalidDicomError
from pydicom.uid import generate_uid


def merge_datasets(datasets: List[Dataset]) -> Dataset:
    new_dataset = copy.copy(datasets[0])

    if len(datasets) > 1:
        new_dataset.SOPInstanceUID = generate_uid()
        new_dataset.file_meta.MediaStorageSOPInstanceUID = generate_uid()
        ordered_by_instance_num = sorted(datasets, key=lambda x: int(x.InstanceNumber))
        frames = [ds.PixelData for ds in ordered_by_instance_num]
        new_dataset.PixelData = b"".join(frames)
        new_dataset.NumberOfFrames = len(frames)

        # TODO Should we delete other fields from `new_dataset`?
        if "InstanceNumber" in new_dataset:
            del new_dataset["InstanceNumber"]

    return new_dataset


def create_merged_datasets(path: Path) -> List[Dataset]:
    series_lookup: Dict[str, Any] = defaultdict(list)

    for file in path.rglob("*"):
        try:
            ds = pydicom.dcmread(file)
            series_lookup[ds.SeriesInstanceUID].append(ds)
        except InvalidDicomError as e:
            logging.exception(f"Failed to open `{file}`", exc_info=e)

    merged_datasets = [merge_datasets(datasets) for datasets in series_lookup.values()]
    return merged_datasets


def merge_series(path: Path) -> None:
    # Sometimes each frame for a 3D image is stored in a separate DICOM.
    # This code merges the separate frame files into a single DICOM file containing the 3D image.
    for i, dataset in enumerate(create_merged_datasets(path)):
        dataset.save_as(path / f"merged-{i}.dcm")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import logging
import multiprocessing
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import partial
from itertools import islice
from os import PathLike
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from registry import RegisteredFunction, Registry
from tqdm_multiprocessing import ConcurrentMapper

from ..dicom import Dicom

# Type checking fails when dataclass attr name matches a type alias.
# Import types under a different alias
from .record import HELPER_REGISTRY, RECORD_REGISTRY, DicomFileRecord, FileRecord, RecordHelper, StandardizedFilename


logger = logging.getLogger(__name__)

FILTER_REGISTRY = Registry("filters")

C = TypeVar("C", bound="RecordCollection")


@dataclass
class RecordFilter:
    r""":class:`RecordFilter` defines a filter function that can be used to filter :class`FileRecord`s during
    the record discovery phase.

    The following filter hooks are available:
        * :func:`path_is_valid` - Checks a path before any :class:`FileRecord` creation is attempted.
        * :func:`record_is_valid` - Checks the created :class:`FileRecord`

    If either of the filter hooks return ``False``, the record will not be included in the iterator of
    discovered :class:`FileRecord`s.
    """

    def __call__(self, target: Union[Path, FileRecord]) -> bool:
        if isinstance(target, Path):
            valid = self.path_is_valid(target)
        elif isinstance(target, FileRecord):
            valid = self.record_is_valid(target)
        else:
            raise TypeError(type(target))
        return valid

    def path_is_valid(self, path: Path) -> bool:
        return True

    def record_is_valid(self, rec: FileRecord) -> bool:
        return True


@dataclass
class CollectionHelper(ABC):
    r"""A :class:`RecordHelper` implements logic that is run during :class:`FileRecord`
    creation and populates fields using custom logic.
    """

    @abstractmethod
    def __call__(self, collection: C, index: int = 0) -> C:
        r"""Applies postprocessing logic to a :class:`FileRecord`.

        Args:
            path:
                Path of the
        """
        return collection


def apply_helpers(collection: C, helpers: Iterable[CollectionHelper], index: int = 0) -> C:
    for h in helpers:
        if not isinstance(h, CollectionHelper):
            raise TypeError(f"type {type(h)} is not a `CollectionHelper`")
        logger.debug(f"Applying collection helper {type(h)}")
        collection = h(cast(C, collection), index)
        if not isinstance(collection, RecordCollection):
            raise TypeError(f"Collection helper {type(h)} returned invalid type {type(collection)}")
    return cast(C, collection)


class RecordCreator:
    r"""RecordCreators examine candidate files and determine which :class:`FileRecord` subclass, if any,
    should be created for the candidate. A :class:`FileRecord` subclass will be chosen by trying callables
    in ``functions`` until one succeeds without raising an exception. Subclasses will be tried before
    their parents.

    Args:
        functions:
            Iterable of names for registered :class:`FileRecord`s to try when creating records.

        helpers:
            Iterable of names for registered :class:`RecordHelper`s to use when creating records.
    """

    def __init__(
        self,
        functions: Optional[Iterable[str]] = None,
        helpers: Iterable[str] = [],
    ):
        functions = RECORD_REGISTRY.available_keys() if functions is None else functions
        self.functions = [RECORD_REGISTRY.get(name) for name in functions]

        self.helpers = []
        for h in helpers:
            helper = HELPER_REGISTRY.get(h).instantiate_with_metadata().fn
            if isinstance(helper, RecordHelper):
                self.helpers.append(helper)
            else:
                logger.debug(
                    f"Ignoring non-RecordHelper `{h}` of type {type(helper)} in RecordCreator. "
                    "It may be still used elsewhere"
                )

    def __call__(self, path: Path, dcm: Optional[Dicom] = None) -> FileRecord:
        result = FileRecord(path)

        for dtype in self.iterate_types_to_try(path):
            try:
                # NOTE: we want to avoid repeatedly opening a DICOM file for each invalid subclass.
                # As such, open `dcm` once with all tags present and reuse it
                if issubclass(dtype, DicomFileRecord):
                    if dcm is None:
                        dcm = dtype.read(path, specific_tags=None, helpers=self.helpers)
                    result = dtype.from_dicom(
                        path,
                        dcm,
                        helpers=self.helpers,
                    )
                else:
                    result = dtype.from_file(path, helpers=self.helpers)
                logger.debug(f"Created {dtype.__name__} for {path}")
                break
            except Exception as ex:
                logger.debug(f"Caught exception creating {dtype} for {path}")
                logger.debug(ex, exc_info=True)

        return result

    def iterate_types_to_try(self, path: PathLike) -> Iterator[Type[FileRecord]]:
        r"""Iterates over the :class:`FileRecord` subclasses to try when creating ``path``
        in the order in which they should be tried.
        """
        path = Path(path)
        candidates: List[Type[FileRecord]] = []

        # exclude candidates by file extension if `path` has an extension
        for fn in self.functions:
            assert isinstance(fn, RegisteredFunction)
            assert isinstance(fn.fn, type)
            assert issubclass(fn.fn, FileRecord)

            # Some DICOMs are named according to UIDs (which include '.') and also have no suffix.
            # We want to ensure we can still identify these as DICOMs even though they technically have
            # an invalid suffix. We also want to avoid checking all files for DICOM headers since this will
            # slow down record creation. As such, we only check valid suffix files that match the expected
            # suffixes, or files with an invalid suffix.
            suffix_is_valid = len(path.suffix) <= 5 and re.search(r"\D", path.suffix.lstrip(".")) is not None
            suffixes = set(fn.metadata.get("suffixes", []))
            path_suffix = path.suffix.lower()
            valid_candidate = not suffix_is_valid or not path_suffix or not suffixes or path_suffix in suffixes

            if valid_candidate:
                candidates.append(fn.fn)

        dicom_candidates = set(self.filter_dicom_types(candidates))
        non_dicom_candidates = set(candidates) - dicom_candidates

        yield from self.sort_by_subclass_level(dicom_candidates)
        yield from self.sort_by_subclass_level(non_dicom_candidates)

    @classmethod
    def sort_by_subclass_level(cls, types: Iterable[Type[FileRecord]]) -> List[Type[FileRecord]]:
        return sorted(types, key=lambda t: len(t.__mro__), reverse=True)

    @classmethod
    def filter_dicom_types(cls, types: Iterable[Type[FileRecord]]) -> List[Type[DicomFileRecord]]:
        return [t for t in types if issubclass(t, DicomFileRecord)]

    @classmethod
    def precheck_file(cls, path: PathLike, filter_funcs: Iterable[Callable[..., bool]]) -> Optional[Path]:
        path = Path(path)
        if path.is_file():
            for f in filter_funcs:
                if not f(path):
                    logger.debug(f"Filter {f} excluded {path}")
                    break
            else:
                return path
        return None


def get_bar_description_suffix(jobs: Optional[int], threads: bool) -> str:
    jobs = jobs or multiprocessing.cpu_count()
    return f"{jobs} {'threads' if threads else 'processes'}"


def record_iterator(
    files: Iterable[PathLike],
    jobs: Optional[int] = None,
    use_bar: bool = True,
    threads: bool = False,
    record_types: Optional[Iterable[str]] = None,
    helpers: Iterable[str] = [],
    ignore_exceptions: bool = False,
    filters: Iterable[str] = [],
    max_files: Optional[int] = None,
    timeout: Optional[int] = None,
    chunksize: int = 128,
    **kwargs,
) -> Iterator[FileRecord]:
    r"""Produces :class:`FileRecord` instances by iterating over an input list of files. If a
    :class:`FileRecord` cannot be from_filed from a path, it will be ignored.

    Args:
        files:
            List of paths to iterate over

        jobs:
            Number of parallel processes to use

        use_bar:
            If ``False``, don't show a tqdm progress bar

        threads:
            If ``True``, use a :class:`ThreadPoolExecutor`. Otherwise, use a :class:`ProcessPoolExecutor`

        record_types:
            List of registered names for :class:`FileRecord` types to try

        helpers:
            Iterable of registered names for :class:`RecordHelper`s to use when creating records

        ignore_exceptions:
            If ``False``, any exceptions raised during record creation will not be suppressed. By default,
            exceptions are silently ignored and records will not be produced for failing files.

        filters:
            Iterable of registered names for :class:`RecordFilter`s to use when filtering potential records

    Keyword Args:
        Forwarded to :class:`RecordCreator`

    Returns:
        Iterator of :class:`FileRecord`s
    """
    # build the RecordCreator which determines what FileRecord subclass to use for each file
    filter_funcs: List[Callable[..., bool]] = [FILTER_REGISTRY.get(f).instantiate_with_metadata() for f in filters]
    creator = RecordCreator(record_types, helpers, **kwargs)
    logger.info("Staring record_iterator")
    logger.info(f"Functions: {creator.functions}")
    logger.info(f"Helpers: {creator.helpers}")
    logger.info(f"Filters: {filter_funcs}")

    # build a list of files to check
    # a record creation attempt will only be made on valid files that pass all filters

    func = partial(creator.precheck_file, filter_funcs=filter_funcs)
    bar_suffix = get_bar_description_suffix(jobs, threads)
    files = files if max_files else islice(files, max_files)
    with ConcurrentMapper(threads, jobs, ignore_exceptions, timeout=timeout, chunksize=chunksize) as mapper:
        mapper.create_bar(
            desc=f"Scanning sources ({bar_suffix})",
            disable=(not use_bar),
        )
        file_set = {Path(path) for path in mapper(func, files) if path is not None}
        mapper.close_bar()

        # create and yield records
        mapper.create_bar(
            desc=f"Building records ({bar_suffix})",
            total=len(file_set),
            unit="file",
            disable=(not use_bar),
        )
        for record in mapper(creator, file_set):
            if all(f(record) for f in filter_funcs):
                yield record


A = TypeVar("A")
T = TypeVar("T", bound=Hashable)


def search_dir(path: PathLike, pattern: str) -> Iterable[Path]:
    path = Path(path)
    if not path.is_dir():
        raise NotADirectoryError(path)
    return path.rglob(pattern)


def iterate_filepaths(
    paths: Iterable[PathLike],
    pattern: str,
    raise_errors: bool = True,
) -> Iterator[Path]:
    for path in paths:
        path = Path(path)
        if path.is_file():
            yield path
        elif path.is_dir():
            for p in search_dir(path, pattern):
                yield p
        elif raise_errors:
            raise FileNotFoundError(path)


def iterate_input_path(
    path: Union[PathLike, Iterable[PathLike]],
    max_depth: Optional[int] = None,
    ignore_missing: bool = False,
    follow_symlinks: bool = True,
    follow_text_files: bool = True,
    _depth: int = 0,
) -> Iterator[Path]:
    r"""Iterates over an input path, yielding all files found.

    Inputs are processed as follows:
        * If the path is a text file and ``follow_text_files`` and we are at depth 0, the text file is assumed to
          contain a list of paths to process. Each line is stripped and processed as a path.
          Invalid paths will raise ``FileNotFoundError`` if ``ignore_missing`` is ``False``.
        * If the path is a file, it will be yielded.
        * If the path is a directory, it will be recursively searched for files up to ``max_depth``.
          If ``max_depth`` is reached, the directory will be yielded instead of recursing further.
        * If the path is a symlink, it will be validated and followed or ignored depending on ``follow_symlinks``.

    Args:
        path:
            Path to process
        max_depth:
            Maximum depth to recurse into directories. If ``None``, recurse infinitely.
        ignore_missing:
            If ``True``, ignore missing files and directories. Otherwise, raise ``FileNotFoundError``.
        follow_symlinks:
            If ``True``, follow symbolic links. Otherwise, ignore them.
        follow_text_files:
            If ``True``, follow symbolic links. Otherwise, ignore them.
    """
    # if input is an interable of paths, recurse over them
    if isinstance(path, Iterable):
        assert _depth == 0, "Cannot iterate over an iterable of paths more than once"
        for p in path:
            yield from iterate_input_path(p, max_depth, ignore_missing, follow_symlinks, follow_text_files, _depth)
        return
    else:
        path = Path(path)

    # if path is a symlink, perform some checks before deferring to other handlers
    if path.is_symlink():
        # maybe ignore it
        if not follow_symlinks:
            return

        # check if its valid
        if not path.resolve().exists():
            if ignore_missing:
                return
            raise FileNotFoundError(f"Symbolic link {path} points to a non-existent file")

    # if path is a directory
    if path.is_dir():
        # possibly recurse into subdirectories
        if max_depth is None or _depth < max_depth:
            for child in path.iterdir():
                yield from iterate_input_path(
                    child, max_depth, ignore_missing, follow_symlinks, follow_text_files, _depth=_depth + 1
                )
        # otherwise yield the directory itself
        else:
            yield path

    # if path is a text file and we are at depth 0, read all lines in the file if requested
    elif path.is_file() and path.suffix == ".txt" and _depth == 0 and follow_text_files:
        with open(path, "r") as f:
            for line in f:
                yield from iterate_input_path(
                    Path(line.strip()), max_depth, ignore_missing, follow_symlinks, follow_text_files, _depth=0
                )

    # if path is any other file, return it
    elif path.is_file():
        yield path

    elif not ignore_missing:
        raise FileNotFoundError(path)


R = TypeVar("R", bound=FileRecord)


class RecordCollection(Generic[R]):
    r"""Data stucture for organizing :class:`FileRecord` instances, indexed by various attriburtes."""

    def __init__(self, records: Iterable[R] = set()):
        self.records: Set[R] = set(records)

    def __repr__(self) -> str:
        types = {}
        for record in self.records:
            name = record.__class__.__name__
            types.setdefault(name, 0)
            types[name] += 1
        root = self.common_parent_dir()
        sample_rec = next(iter(self.records), None)
        sample_path = sample_rec.path if sample_rec else None
        return f"{self.__class__.__name__}(length={len(self)}, types={types}, parent={root}, sample_path={sample_path})"

    def __len__(self) -> int:
        r"""Returns the total number of contained records"""
        return len(self.records)

    def __iter__(self) -> Iterator[R]:
        for rec in self.records:
            yield rec

    def __contains__(self, x: Any) -> bool:
        return x in self.records

    def add(self, rec: R) -> None:
        return self.records.add(rec)

    def pop(self) -> R:
        return self.records.pop()

    def union(self: C, other: C) -> C:
        return self.__class__(self.records.union(other.records))

    def intersection(self: C, other: C) -> C:
        return self.__class__(self.records.intersection(other.records))

    def difference(self: C, other: C) -> C:
        return self.__class__(self.records.difference(other.records))

    def apply(self: C, func: Callable[[R], R]) -> C:
        records = {func(rec) for rec in self.records}
        return self.__class__(records)

    def filter(self: C, func: Callable[[R], bool]) -> C:
        records = {rec for rec in self.records if func(rec)}
        return self.__class__(records)

    def contains_record_type(self, dtype: Type[FileRecord]) -> bool:
        return any(isinstance(rec, dtype) for rec in self)

    @overload
    def group_by(self: C, func: Callable[[R], T]) -> Dict[T, C]: ...

    @overload
    def group_by(self: C, func: Callable[[R], T], *funcs: Callable[[R], T]) -> Dict[Tuple[T, ...], C]: ...

    def group_by(
        self: C, func: Callable[[R], T], *funcs: Callable[[R], T]
    ) -> Union[Dict[T, C], Dict[Tuple[T, ...], C]]:
        result: Dict[T, C] = {}
        funcs = (func, *funcs)
        for record in self.records:
            key = tuple(f(record) for f in funcs)
            if len(funcs) == 1:
                key = key[0]
            result.setdefault(cast(T, key), self.__class__()).add(record)
        return result

    def parent_dirs(self, offset: int = 0) -> Set[Path]:
        r"""Gets the set of unique parent directories at level ``offset``."""
        return {rec.path.parents[offset] for rec in self.records}

    def common_parent_dir(self) -> Optional[Path]:
        parents = self.parent_dirs()
        if not parents:
            return None
        proto = next(iter(parents))
        while proto.parent != proto:
            if all(p.is_relative_to(proto) for p in parents):
                return proto
            proto = proto.parent
        return None

    def standardized_filenames(self) -> Iterator[Tuple[StandardizedFilename, R]]:
        counter: Dict[str, int] = {}
        for rec in self:
            filename = rec.standardized_filename("id")
            prefix = filename.prefix
            count = counter.get(prefix, 1)
            filename = filename.with_file_id(count)
            yield filename, rec
            counter[prefix] = count + 1

    def prune_duplicates(self: C, func: Callable[[R], T]) -> C:
        groups = self.group_by(func)
        return self.__class__({next(iter(g)) for g in groups.values()})

    def to_dict(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        result["records"] = {}
        for name, record in self.standardized_filenames():
            result["records"][str(name)] = record.to_dict()
        return result

    def to_json(self, path: PathLike, indent: int = 4) -> None:
        path = Path(path)
        with open(path, "w") as f:
            json.dump(self.to_dict(), f, sort_keys=True, indent=indent)

    @classmethod
    def from_dir(
        cls: Type[C],
        path: PathLike,
        pattern: str = "*",
        jobs: Optional[int] = None,
        use_bar: bool = True,
        threads: bool = False,
        record_types: Optional[Iterable[str]] = None,
        helpers: Iterable[str] = [],
        filters: Iterable[str] = [],
        **kwargs,
    ) -> C:
        r"""Create a :class:`RecordCollection` from files in a directory matching a wildcard.
        If a :class:`FileRecord` cannot be created for a file, that file is silently excluded
        from the collection.

        Args:
            path:
                Path to the directory to search

            pattern:
                Glob pattern for matching files

            jobs:
                Number of parallel jobs to use

            use_bar:
                If ``False``, don't show a tqdm progress bar

            threads:
                If ``True``, use a :class:`ThreadPoolExecutor`. Otherwise, use a :class:`ProcessPoolExecutor`

            record_types:
                List of registered names for :class:`FileRecord` types to try

            helpers:
                Iterable of registered names for :class:`RecordHelper`s to use when creating records

            filters:
                Iterable of registered names for :class:`RecordFilter`s to use when filtering potential records

        Keyword Args:
            Forwarded to :func:`record_iterator`
        """
        return cls.from_files(
            search_dir(path, pattern), jobs, use_bar, threads, record_types, helpers, filters, **kwargs
        )

    @classmethod
    def from_files(
        cls: Type[C],
        files: Iterable[PathLike],
        jobs: Optional[int] = None,
        use_bar: bool = True,
        threads: bool = False,
        record_types: Optional[Iterable[str]] = None,
        helpers: Iterable[str] = [],
        filters: Iterable[str] = [],
        **kwargs,
    ) -> C:
        r"""Create a :class:`RecordCollection` from a list of files.
        If a :class:`FileRecord` cannot be created for a file, that file is silently excluded
        from the collection.

        Args:
            files:
                List of files to create records from

            jobs:
                Number of parallel jobs to use

            use_bar:
                If ``False``, don't show a tqdm progress bar

            threads:
                If ``True``, use a :class:`ThreadPoolExecutor`. Otherwise, use a :class:`ProcessPoolExecutor`

            record_types:
                List of registered names for :class:`FileRecord` types to try

            helpers:
                Iterable of registered names for :class:`RecordHelper`s to use when creating records

            filters:
                Iterable of registered names for :class:`RecordFilter`s to use when filtering potential records

        Keyword Args:
            Forwarded to :func:`record_iterator`
        """
        collection = cls(
            record_iterator(files, jobs, use_bar, threads, record_types, helpers, filters=filters, **kwargs)
        )
        return collection

    @classmethod
    def create(
        cls: Type[C],
        paths: Iterable[PathLike],
        pattern: str = "*",
        jobs: Optional[int] = None,
        use_bar: bool = True,
        threads: bool = False,
        record_types: Optional[Iterable[str]] = None,
        helpers: Iterable[str] = [],
        filters: Iterable[str] = [],
        iterator_kwargs: Dict[str, Any] = {},
        **kwargs,
    ) -> C:
        r"""Create a :class:`RecordCollection` from a list of paths, either files or directories.
        If a :class:`FileRecord` cannot be created for a file, that file is silently excluded
        from the collection.

        Args:
            paths:
                List of paths to create records from

            pattern:
                Glob pattern for matching files

            jobs:
                Number of parallel jobs to use

            use_bar:
                If ``False``, don't show a tqdm progress bar

            threads:
                If ``True``, use a :class:`ThreadPoolExecutor`. Otherwise, use a :class:`ProcessPoolExecutor`

            record_types:
                List of registered names for :class:`FileRecord` types to try

            helpers:
                Iterable of registered names for :class:`RecordHelper`s to use when creating records

            filters:
                Iterable of registered names for :class:`RecordFilter`s to use when filtering potential records

        Keyword Args:
            Forwarded to :func:`record_iterator`
        """
        paths = filter(lambda x: cast(Path, x).match(pattern), iterate_input_path(paths, **iterator_kwargs))
        return cls.from_files(paths, jobs, use_bar, threads, record_types, helpers, filters=filters, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os import PathLike
from pathlib import Path
from typing import Optional, cast

import pydicom
import pytest

from dicom_utils.container import (
    FILTER_REGISTRY,
    HELPER_REGISTRY,
    SOPUID,
    DicomFileRecord,
    DicomImageFileRecord,
    FileRecord,
    MammogramFileRecord,
    RecordCollection,
    RecordCreator,
    RecordFilter,
    RecordHelper,
    StandardizedFilename,
    StudyUID,
    record_iterator,
)
from dicom_utils.container.collection import iterate_input_path
from dicom_utils.dicom_factory import DicomFactory


class TestIterateInputPath:
    def test_file(self, tmp_path):
        path = Path(tmp_path, "test.dcm")
        path.touch()
        result = list(iterate_input_path(path))
        assert result == [path]

    @pytest.mark.parametrize("max_depth", [None, 0, 1])
    def test_dir(self, tmp_path, max_depth):
        path = Path(tmp_path, "dir")
        path.mkdir()
        subfile = Path(path, "test.dcm")
        subfile.touch()
        result = list(iterate_input_path(path, max_depth))
        if max_depth == 0:
            assert result == [path]
        else:
            assert result == list(path.iterdir())

    @pytest.mark.parametrize("follow_text_files", [True, False])
    def test_text_file_of_files(self, tmp_path, follow_text_files):
        path = Path(tmp_path, "test.txt")
        subdir = Path(tmp_path, "dir")
        subdir.mkdir()
        with path.open("w") as f:
            targets = []
            for i in range(3):
                p = Path(subdir, f"test{i}.dcm")
                p.touch()
                targets.append(p)
                f.write(f"{p}\n")
        result = list(iterate_input_path(path, follow_text_files=follow_text_files))
        assert result == (targets if follow_text_files else [path])

    @pytest.mark.parametrize("max_depth", [None, 0, 1])
    def test_text_file_of_dirs(self, tmp_path, max_depth):
        path = Path(tmp_path, "test.txt")
        subdir = Path(tmp_path, "dir")
        subdir.mkdir()
        with path.open("w") as f:
            dir_targets = []
            file_targets = []
            for i in range(3):
                subsubdir = Path(subdir, f"subsubdir{i}")
                subsubdir.mkdir()
                p = Path(subsubdir, f"test{i}.txt")
                # write dummy file content to ensure we don't recursively read text files
                with p.open("w") as f2:
                    f2.write("nofile.txt")
                dir_targets.append(subsubdir)
                file_targets.append(p)
                f.write(f"{subsubdir}\n")
        result = list(iterate_input_path(path, max_depth))
        if max_depth == 0:
            assert result == dir_targets
        else:
            assert result == file_targets

    @pytest.mark.parametrize(
        "ignore_missing",
        [
            True,
            pytest.param(False, marks=pytest.mark.xfail(raises=FileNotFoundError)),
        ],
    )
    def test_missing_filepath(self, tmp_path, ignore_missing):
        path = Path(tmp_path, "test.txt")
        subdir = Path(tmp_path, "dir")
        subdir.mkdir()
        with path.open("w") as f:
            p = Path(subdir, "nofile.txt")
            f.write(f"{p}\n")
        result = list(iterate_input_path(path, ignore_missing=ignore_missing))
        assert result == []

    @pytest.mark.parametrize("follow_symlinks", [True, False])
    def test_handle_symlink_files(self, tmp_path, follow_symlinks):
        real_path = tmp_path / "real"
        symlink_path = tmp_path / "symlink"
        real_path.mkdir()
        symlink_path.mkdir()

        for i in range(3):
            p = real_path / f"test{i}.dcm"
            p.touch()
            symlink = symlink_path / f"test{i}.dcm"
            symlink.symlink_to(p)
            assert symlink.is_file()

        result = list(iterate_input_path(symlink_path, follow_symlinks=follow_symlinks))
        assert len(result) == (3 if follow_symlinks else 0)
        assert all(r.is_symlink() for r in result)

    @pytest.mark.parametrize("follow_symlinks", [True, False])
    def test_handle_symlink_dirs(self, tmp_path, follow_symlinks):
        real_path = tmp_path / "real"
        symlink_path = tmp_path / "symlink"
        real_path.mkdir()
        symlink_path.mkdir()

        for i in range(3):
            p = real_path / str(i) / f"test{i}.dcm"
            p.parent.mkdir()
            p.touch()
            symlink = symlink_path / str(i)
            symlink.symlink_to(p.parent)
            assert symlink.is_dir()

        result = list(iterate_input_path(symlink_path, follow_symlinks=follow_symlinks))
        assert len(result) == (3 if follow_symlinks else 0)
        assert all(r.parent.is_symlink() for r in result)


@pytest.fixture
def suffix():
    return ".dcm"


@pytest.fixture
def dicom_files(tmp_path, dicom_file, suffix):
    paths = []
    factory = DicomFactory(proto=dicom_file)
    for i in range(3):
        for j in range(3):
            study_uid = f"study_{i}"
            sop_uid = f"study_{i}_sop_{j}"
            dest = Path(tmp_path, f"subdir_{i}", f"file_{j}{suffix}")
            dest.parent.mkdir(exist_ok=True, parents=True)

            dcm = factory(StudyInstanceUID=study_uid, SOPInstanceUID=sop_uid)
            dcm.save_as(dest)
            paths.append(dest)
    return paths


class TestRecordCreator:
    @pytest.fixture
    def dicom_factory(self, tmp_path, dicom_file):
        def func(
            filename: PathLike = Path("foo.dcm"),
            Rows: Optional[int] = 128,
            Columns: Optional[int] = 128,
            Modality: Optional[str] = "MG",
            **kwargs,
        ):
            filename = Path(tmp_path, filename)
            filename.parent.mkdir(exist_ok=True, parents=True)

            with pydicom.dcmread(dicom_file) as dcm:
                dcm.Rows = Rows
                dcm.Columns = Columns
                dcm.Modality = Modality
                dcm.save_as(filename)
            return filename

        return func

    @pytest.mark.parametrize(
        "rows,columns,modality,exp",
        [
            pytest.param(128, 128, "MG", MammogramFileRecord),
            pytest.param(128, 128, "CT", DicomImageFileRecord),
            pytest.param(128, 128, "", DicomImageFileRecord),
            pytest.param(None, 128, "MG", DicomFileRecord),
            pytest.param(128, None, "CT", DicomFileRecord),
        ],
    )
    def test_create(self, rows, columns, modality, exp, dicom_factory):
        c = RecordCreator()
        rec = c(dicom_factory(Rows=rows, Columns=columns, Modality=modality))
        assert type(rec) == exp

    @pytest.mark.parametrize("suffix", [".txt", ".html", ".json"])
    def test_create_other_files(self, tmp_path, suffix):
        f = Path(tmp_path, "file").with_suffix(suffix)
        f.touch()
        c = RecordCreator()
        rec = c(f)
        assert type(rec) == FileRecord
        assert rec.path == f


class TestRecordIterator:
    @pytest.mark.parametrize("suffix", [".dcm", "", ".123"])
    @pytest.mark.parametrize("use_bar", [True, False])
    @pytest.mark.parametrize("threads", [False, True])
    @pytest.mark.parametrize("jobs", [None, 1, 2])
    def test_default(self, dicom_files, use_bar, threads, jobs):
        records = list(record_iterator(dicom_files, jobs, use_bar, threads, ignore_exceptions=False))
        assert all(isinstance(r, DicomFileRecord) for r in records)
        assert set(rec.path for rec in records) == set(dicom_files)

    def test_helpers(self, dicom_files):
        @HELPER_REGISTRY(name="dummy-pid")
        class PatientIDHelper(RecordHelper):
            def __call__(self, _, rec):
                return rec.replace(PatientID="TEST")

        records = list(record_iterator(dicom_files, helpers=["dummy-pid"], threads=True))
        assert all(isinstance(r, DicomFileRecord) for r in records)
        assert all(cast(DicomFileRecord, r).PatientID == "TEST" for r in records)

    def test_filters(self, dicom_files):
        @FILTER_REGISTRY(name="dummy-filter")
        class DummyFilter(RecordFilter):
            def path_is_valid(self, path: Path) -> bool:
                return str(path).endswith("0.dcm")

            def record_is_valid(self, rec: FileRecord) -> bool:
                return str(rec.path.parent).endswith("0")

        records = list(record_iterator(dicom_files, filters=["dummy-filter"], threads=True))
        assert all(isinstance(r, DicomFileRecord) for r in records)
        assert all(str(r.path).endswith("0.dcm") for r in records), "path_is_valid filter failed"
        assert all(str(r.path.parent).endswith("0") for r in records), "record_is_valid filter failed"


class TestRecordCollection:
    @pytest.mark.parametrize("use_bar", [True, False])
    @pytest.mark.parametrize("threads", [False, True])
    @pytest.mark.parametrize("jobs", [None, 1, 2])
    def test_from_files(self, dicom_files, use_bar, threads, jobs, caplog):
        col = RecordCollection.from_files(dicom_files, jobs, use_bar, threads)
        assert set(x.path for x in col) == set(dicom_files)
        assert all(isinstance(v, FileRecord) for v in col)
        assert set(v.path for v in col) == set(dicom_files)

    @pytest.mark.parametrize("use_bar", [True, False])
    @pytest.mark.parametrize("threads", [False, True])
    @pytest.mark.parametrize("jobs", [None, 1, 2])
    def test_from_dir(self, tmp_path, dicom_files, use_bar, threads, jobs):
        col = RecordCollection.from_dir(tmp_path, "*.dcm", jobs, use_bar, threads)
        assert set(x.path for x in col) == set(dicom_files)
        assert all(isinstance(v, FileRecord) for v in col)
        assert set(v.path for v in col) == set(dicom_files)

    @pytest.mark.parametrize("use_bar", [True, False])
    @pytest.mark.parametrize("threads", [False, True])
    @pytest.mark.parametrize("jobs", [None, 1, 2])
    def test_create(self, tmp_path, dicom_files, use_bar, threads, jobs):
        files = [tmp_path] + dicom_files
        col = RecordCollection.create(files, "*.dcm", jobs, use_bar, threads)
        assert set(x.path for x in col) == set(dicom_files)
        assert all(isinstance(v, FileRecord) for v in col)
        assert set(v.path for v in col) == set(dicom_files)

    @pytest.fixture
    def collection(self, dicom_files):
        return RecordCollection.from_files(dicom_files)

    def test_len(self, collection, dicom_files):
        assert len(collection) == len(dicom_files)

    def test_standardized_filenames(self, tmp_path, dicom_files):
        col = RecordCollection.from_dir(tmp_path, "*.dcm")
        pairs = list(col.standardized_filenames())
        names = [p[0] for p in pairs]
        assert all(isinstance(n, StandardizedFilename) for n in names)
        assert len(names) == len(col)
        assert len(set(names)) == len(names)

    def test_to_dict(self, collection, dicom_files):
        col_dict = collection.to_dict()
        assert isinstance(col_dict, dict)
        assert len(col_dict["records"]) == len(dicom_files)

    def test_unique(self):
        rec1 = DicomFileRecord(path=Path("foo.dcm"), SOPInstanceUID=SOPUID("123"), StudyInstanceUID=StudyUID("123"))
        rec2 = DicomFileRecord(path=Path("bar.dcm"), SOPInstanceUID=SOPUID("123"), StudyInstanceUID=StudyUID("123"))
        col = RecordCollection([rec1, rec2])
        assert len(col) == 1

    def test_repr(self, tmp_path, collection):
        result = repr(collection)
        expected_start = f"RecordCollection(length=9, types={{'DicomImageFileRecord': 9}}, parent={tmp_path}"
        assert result.startswith(expected_start)

    def test_repr_empty(self):
        result = repr(RecordCollection())
        expected_start = "RecordCollection(length=0, types={}, parent=None"
        assert result.startswith(expected_start)

    @pytest.mark.parametrize(
        "key",
        [
            lambda rec: rec.StudyInstanceUID,
            lambda rec: (rec.StudyInstanceUID, rec.SOPInstanceUID),
        ],
    )
    def test_group_by(self, collection, key):
        grouped = collection.group_by(key)
        assert isinstance(grouped, dict)
        assert all(isinstance(v, RecordCollection) for v in grouped.values())
        assert all(grouped.keys())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
from argparse import ArgumentParser

from ..find_phi import find_phi


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path", help="path to a DICOM file or folder with DICOM files")
    parser.add_argument(
        "--overwrite", help="overwrite original files with anonymized files", default=False, action="store_true"
    )
    parser.add_argument("--verbose", "-v", help="verbose printing", default=False, action="store_true")
    return parser


def main(args: argparse.Namespace) -> None:
    print(f"Overwrite is set to {args.overwrite}")
    print(f"Verbose is set to {args.verbose}")
    find_phi(args.path, args.overwrite, args.verbose)


def entrypoint() -> None:
    main(get_parser().parse_args())


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
from argparse import ArgumentParser
from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from typing import List, Optional

import pydicom

from ..dicom import has_dicm_prefix
from ..types import MammogramType


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path", default="./", help="path to find from")
    parser.add_argument("--name", "-n", default="*", help="glob pattern for filename")
    parser.add_argument("--parents", "-p", default=False, action="store_true", help="return unique parent directories")
    parser.add_argument(
        "--types", "-t", choices=["ffdm", "sfm", "tomo", "s-view"], default=None, nargs="+", help="filter by image type"
    )
    parser.add_argument("--jobs", "-j", default=4, help="parallelism")
    # TODO add a field to only return DICOMs with readable image data
    return parser


def is_desired_type(path: Path, types: List[str]) -> bool:
    with pydicom.dcmread(path, stop_before_pixels=True) as dcm:
        mammogram_type = MammogramType.from_dicom(dcm)
    return str(mammogram_type) in types


def check_file(path: Path, args: argparse.Namespace) -> Optional[Path]:
    if not path.is_file() or not has_dicm_prefix(path):
        return
    try:
        if args.types is not None and not is_desired_type(path, args.types):
            return
    except Exception:
        return

    return path


def main(args: argparse.Namespace) -> None:
    path = Path(args.path)
    if not path.is_dir():
        raise NotADirectoryError(path)

    seen_parents = set()

    def callback(x: Future):
        result = x.result()
        if result is None:
            return

        try:
            if args.parents and result.parent not in seen_parents:
                print(result.parent, flush=True)
            elif not args.parents:
                print(result, flush=True)
            seen_parents.add(result.parent)
        except IOError:
            tp.shutdown(wait=False)

    futures: List[Future] = []
    with ThreadPoolExecutor(args.jobs) as tp:
        for match in path.rglob(args.name):
            f = tp.submit(check_file, match, args)
            futures.append(f)

        for f in futures:
            callback(f)


def entrypoint():
    parser = get_parser()
    main(parser.parse_args())


if __name__ == "__main__":
    entrypoint()



import logging
from enum import Enum
from typing import Optional, Union

from pydicom.config import logger as pydicom_logger


logging.basicConfig()
logger = logging.getLogger("dicom-utils")


class ExtendedEnum(Enum):
    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class LoggingLevel(ExtendedEnum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


Level = Union[LoggingLevel, str]


def set_logging_level(logging_level: Level, pydicom_logging_level: Optional[Level]) -> None:
    """
    Set logging levels for this package.
    If pydicom_logging_level is None, logging_level will override pydicom_logging_level.

    Args:
        logging_level:
            Logging level for this package which may override the pydicom logging level
        pydicom_logging_level:
            Logging level for pydicom

    Returns:
        None
    """
    logging_level = LoggingLevel(logging_level)
    pydicom_logging_level = LoggingLevel(pydicom_logging_level or logging_level)

    logger.setLevel(logging_level.value)
    pydicom_logger.setLevel(pydicom_logging_level.value)

    logger.debug(f"Package logging set to {logging_level} and pydicom logging set to {pydicom_logging_level}.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import logging
import warnings
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import Callable, Dict, Final, Hashable, Iterator, Optional, Sequence, Tuple, TypeVar, cast

from registry import Registry
from tqdm_multiprocessing import ConcurrentMapper

from .collection import CollectionHelper, RecordCollection, apply_helpers, get_bar_description_suffix

# Type checking fails when dataclass attr name matches a type alias.
# Import types under a different alias
from .helpers import StudyUID
from .record import HELPER_REGISTRY, FileRecord


H = TypeVar("H", bound=Hashable)


GroupFunction = Callable[[FileRecord], Hashable]
GROUP_REGISTRY = Registry("group", bound=Callable[..., Hashable])
Key = Tuple[Hashable, ...]
GroupDict = Dict[Key, RecordCollection]
COMPLETE_COLLECTION_INDEX: Final[int] = 0
FIRST_GROUP_INDEX: Final[int] = 1

logger = logging.getLogger(__name__)


@GROUP_REGISTRY(name="study-uid")
def group_by_study_instance_uid(rec: FileRecord) -> Optional[StudyUID]:
    return getattr(rec, "StudyInstanceUID", None)


@GROUP_REGISTRY(name="parent")
def group_by_parent(rec: FileRecord, level: int = 0) -> Path:
    return rec.path.parents[level]


for i in range(3):
    GROUP_REGISTRY(partial(group_by_parent, level=i + 1), name=f"parent-{i + 1}")


@GROUP_REGISTRY(name="patient-id")
def group_by_patient_id(rec: FileRecord) -> Optional[str]:
    return getattr(rec, "PatientID", None)


@GROUP_REGISTRY(name="study-date")
def group_by_study_date(rec: FileRecord) -> Optional[str]:
    return getattr(rec, "StudyDate", None)


@dataclass
class Grouper:
    groups: Sequence[str]
    helpers: Sequence[str]
    threads: bool = False
    jobs: Optional[int] = None
    chunksize: int = 8
    use_bar: bool = True
    timeout: Optional[int] = None

    _group_fns: Sequence[GroupFunction] = field(init=False)
    _helper_fns: Sequence[CollectionHelper] = field(init=False)

    def __post_init__(self):
        if not self.groups:
            raise ValueError("`groups` cannot be empty")

        self._group_fns = [GROUP_REGISTRY.get(g).instantiate_with_metadata().fn for g in self.groups]
        self._helper_fns = list(self._build_collection_helpers())
        logger.info(f"Collection helpers: {self._helper_fns}")

        # Using processes seems to result in deadlocks. This is a workaround.
        if not self.threads:
            self.threads = True

    def _build_collection_helpers(self) -> Iterator[CollectionHelper]:
        for h in self.helpers:
            helper = HELPER_REGISTRY.get(h).instantiate_with_metadata().fn
            # isolate CollectionHelpers
            if isinstance(helper, CollectionHelper):
                yield helper
            else:
                logger.debug(f"Ignoring non-RecordHelper `{h}` of type {type(h)}")

    def __call__(self, collection: RecordCollection) -> Dict[Hashable, RecordCollection]:
        start_len = len(collection)

        # Apply helpers at the entire collection level
        collection = apply_helpers(collection, self._helper_fns, index=COMPLETE_COLLECTION_INDEX)

        result: Dict[Key, RecordCollection] = {tuple(): collection}

        with ConcurrentMapper(self.threads, self.jobs, chunksize=self.chunksize, timeout=self.timeout) as mapper:
            for i, group_fn in list(enumerate(self._group_fns)):
                # run the group function
                total = sum(len(v) for v in result.values())
                mapper.create_bar(
                    desc=f"Running grouper {self.groups[i]} ({get_bar_description_suffix(self.jobs, self.threads)})",
                    disable=(not self.use_bar),
                    leave=True,
                    total=total,
                )
                _result: Dict[Key, RecordCollection] = {}
                needs_grouping: Iterator[Tuple[Key, FileRecord]] = (
                    (key, record) for key, collection in result.items() for record in collection
                )
                # TODO: This can deadlock. Only seems to happen on Optimam full dataset.
                for key, record in mapper(self._group, needs_grouping, group_fn=group_fn):
                    _result.setdefault(key, RecordCollection()).add(record)
                result = _result
                mapper.close_bar()

                # run helpers for this stage in the grouping process
                mapper.create_bar(
                    desc=f"Running group helpers ({get_bar_description_suffix(self.jobs, self.threads)})",
                    disable=(not self.use_bar),
                    leave=True,
                    total=len(result),
                )
                mapped = mapper(self._apply_helpers, list(result.items()), index=i + 1)
                result = {k: v for k, v in mapped}
                mapper.close_bar()

        end_len = sum(len(collection) for collection in result.values())
        if start_len != end_len:
            warnings.warn(
                f"Grouping began with {start_len} records and ended with {end_len} records. "
                "This may be expected behavior depending on what helpers are being used."
            )
        return cast(Dict[Hashable, RecordCollection], result)

    @classmethod
    def _group(cls, inp: Tuple[Key, FileRecord], group_fn: GroupFunction) -> Tuple[Key, FileRecord]:
        key, entry = inp
        key = key + (group_fn(entry),)
        return key, entry

    def _apply_helpers(self, inp: Tuple[Key, RecordCollection], index: int) -> Tuple[Key, RecordCollection]:
        k, col = inp
        return k, apply_helpers(col, self._helper_fns, index=index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict, cast

import pytest
from pydicom import DataElement
from pydicom.multival import MultiValue

from dicom_utils import DicomFactory
from dicom_utils.tags import Tag
from dicom_utils.types import (
    ImageType,
    Laterality,
    MammogramType,
    MammogramView,
    PhotometricInterpretation,
    PixelSpacing,
    ViewPosition,
)


@dataclass
class DummyElement:
    value: Any


def get_simple_image_type_test_cases():
    """Seen IMAGE_TYPE Fields:

    2D:
        ['ORIGINAL', 'PRIMARY']
        ['DERIVED', 'PRIMARY']
        ['ORIGINAL', 'PRIMARY', '', '', '', '', '', '', '150000']
        ['DERIVED', 'PRIMARY', 'POST_CONTRAST', 'SUBTRACTION', '', '', '', '', '50000']
        ['ORIGINAL', 'PRIMARY', 'POST_PROCESSED', '', '', '', '', '', '50000']
        ['DERIVED', 'PRIMARY', 'TOMO_PROJ', 'RIGHT', '', '', '', '', '150000'] (may be s-view, but no marker on image)
        ['DERIVED', 'SECONDARY']
        ['DERIVED', 'PRIMARY', 'TOMO_2D', 'LEFT', '', '', '', '', '150000']
        ['DERIVED', 'PRIMARY', 'TOMO_2D', 'RIGHT', '', '', '', '', '150000']

    S-View:
        ['DERIVED', 'PRIMARY', '', '', '', '', '', '', '150000']
        ['DERIVED', 'PRIMARY', 'TOMO', 'GENERATED_2D', '', '', '', '', '150000']
        ['DERIVED', 'PRIMARY', 'TOMOSYNTHESIS', 'GENERATED_2D', '', '', '', '', '150000']

    TOMO:
        ['DERIVED', 'PRIMARY', 'TOMOSYNTHESIS', 'NONE', '', '', '', '', '150000']

    """
    cases = []
    default: Dict[str, Any] = {"pixels": "ORIGINAL", "exam": "PRIMARY"}

    # 2D

    # ['ORIGINAL', 'PRIMARY']
    d = deepcopy(default)
    _ = pytest.param(d, MammogramType.FFDM, id="2d-1")
    cases.append(_)

    # ['DERIVED', 'PRIMARY']
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED"))
    _ = pytest.param(d, MammogramType.FFDM, id="2d-2")
    cases.append(_)

    # ['ORIGINAL', 'PRIMARY', '', '', '', '', '', '', '150000']
    d = deepcopy(default)
    d.update(dict(pixels="ORIGINAL", extras=["", "", "", "", "", "150000"]))
    _ = pytest.param(d, MammogramType.FFDM, id="2d-3")
    cases.append(_)

    # ['DERIVED', 'PRIMARY', 'POST_CONTRAST', 'SUBTRACTION', '', '', '', '', '50000']
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", flavor="POST_CONTRAST", extras=["SUBTRACTION", "", "", "50000"]))
    _ = pytest.param(d, MammogramType.FFDM, id="2d-4")
    cases.append(_)

    # ['ORIGINAL', 'PRIMARY', 'POST_PROCESSED', '', '', '', '', '', '50000']
    d = deepcopy(default)
    d.update(dict(pixels="ORIGINAL", flavor="POST_PROCESSED", extras=["", "", "50000"]))
    _ = pytest.param(d, MammogramType.FFDM, id="2d-5")
    cases.append(_)

    # ['DERIVED', 'PRIMARY', 'TOMO_PROJ', 'RIGHT', '', '', '', '', '150000'] (may be s-view, but no marker on image)
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", flavor="TOMO_PROJ", extras=["RIGHT", "", "50000"]))
    _ = pytest.param(d, MammogramType.FFDM, id="2d-6")
    cases.append(_)

    # ['DERIVED', 'SECONDARY']
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", exam="SECONDARY"))
    _ = pytest.param(d, MammogramType.FFDM, id="2d-7")
    cases.append(_)

    # ['DERIVED', 'PRIMARY', 'TOMO_2D', 'LEFT', '', '', '', '', '150000']
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", flavor="TOMO_2D", extras=["LEFT", "", "", "", "150000"]))
    _ = pytest.param(d, MammogramType.FFDM, id="2d-8")
    cases.append(_)

    # ['DERIVED', 'PRIMARY', 'TOMO_2D', 'RIGHT', '', '', '', '', '150000']
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", flavor="TOMO_2D", extras=["RIGHT", "", "", "", "150000"]))
    _ = pytest.param(d, MammogramType.FFDM, id="2d-9")
    cases.append(_)

    # S-VIEW

    # ['DERIVED', 'PRIMARY', 'TOMO', 'GENERATED_2D', '', '', '', '', '150000']
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", flavor="TOMO", extras=["GENERATED_2D", "", "", "", "150000"]))
    _ = pytest.param(d, MammogramType.SYNTH, id="sview-1")
    cases.append(_)

    # ['DERIVED', 'PRIMARY', 'TOMOSYNTHESIS', 'GENERATED_2D', '', '', '', '', '150000']
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", flavor="TOMOSYNTHESIS", extras=["GENERATED_2D", "", "", "", "150000"]))
    _ = pytest.param(d, MammogramType.SYNTH, id="sview-2")
    cases.append(_)

    # ['DERIVED', 'PRIMARY']
    # Data in SeriesDescription
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", series_description="L CC C-View"))
    _ = pytest.param(d, MammogramType.SYNTH, id="sview-3")
    cases.append(_)

    # ['DERIVED', 'PRIMARY']
    # Data in SeriesDescription
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", series_description="R MLO S-View"))
    _ = pytest.param(d, MammogramType.SYNTH, id="sview-4")
    cases.append(_)

    # TOMO

    # ['DERIVED', 'PRIMARY', 'TOMOSYNTHESIS', 'NONE', '', '', '', '', '150000']
    d = deepcopy(default)
    d.update(dict(pixels="DERIVED", NumberOfFrames=10, flavor="TOMOSYNTHESIS", extras=["NONE", "", "", "150000"]))
    _ = pytest.param(d, MammogramType.TOMO, id="tomo-1")
    cases.append(_)

    return cases


class TestImageType:
    @pytest.mark.parametrize(
        "pixels,exam,flavor,extras",
        [
            ("ORIGINAL", "PRIMARY", None, None),
            ("DERIVED", "PRIMARY", "AXIAL", None),
            ("DERIVED", "PRIMARY", "AXIAL", ["STEREO"]),
        ],
    )
    def test_from_dicom(self, dicom_object, pixels, exam, flavor, extras):
        dicom_object.ImageType = [x for x in (pixels, exam, flavor, *(extras or [])) if x]
        img_type = ImageType.from_dicom(dicom_object)
        assert img_type.pixels == pixels
        assert img_type.exam == exam
        assert img_type.flavor == flavor
        assert img_type.extras == extras

    @pytest.mark.parametrize(
        "pixels,exam,flavor,extras,target,expected",
        [
            ("ORIGINAL", "PRIMARY", "AXIAL", [], "ORIGINAL", True),
            ("ORIGINAL", "PRIMARY", "AXIAL", [], "PRIMARY", True),
            ("ORIGINAL", "PRIMARY", "AXIAL", [], "AXIAL", True),
            ("ORIGINAL", "PRIMARY", None, None, "AXIAL", False),
            ("ORIGINAL", "PRIMARY", None, ["AXIAL"], "AXIAL", True),
        ],
    )
    def test_contains(self, pixels, exam, flavor, extras, target, expected):
        img_type = ImageType(pixels, exam, flavor, extras)
        assert (target in img_type) == expected


class TestMammogramType:
    @pytest.mark.parametrize(
        "input_str, expected",
        [
            ("ffdm", MammogramType.FFDM),
            ("2d", MammogramType.FFDM),
            ("synth", MammogramType.SYNTH),
            ("s view", MammogramType.SYNTH),
            ("s-view", MammogramType.SYNTH),
            ("c-view", MammogramType.SYNTH),
            ("", MammogramType.UNKNOWN),
            ("unknown", MammogramType.UNKNOWN),
            ("tomo", MammogramType.TOMO),
            ("tomosynthesis", MammogramType.TOMO),
        ],
    )
    def test_from_str(self, input_str, expected):
        assert expected == MammogramType.from_str(input_str)

    @pytest.mark.parametrize(
        "val,exp",
        [
            (MammogramType.UNKNOWN, False),
            *[(val, True) for val in MammogramType if val != MammogramType.UNKNOWN],
        ],
    )
    def test_bool(self, val, exp):
        assert bool(val) == exp

    @pytest.mark.parametrize(
        "val,exp",
        [
            (MammogramType.UNKNOWN, True),
            *[(val, False) for val in MammogramType if val != MammogramType.UNKNOWN],
        ],
    )
    def test_is_unknown(self, val, exp):
        assert val.is_unknown == exp

    @pytest.mark.parametrize(
        "t1,t2,exp",
        [
            (MammogramType.FFDM, MammogramType.FFDM, False),
            (MammogramType.FFDM, MammogramType.SYNTH, True),
            (MammogramType.FFDM, MammogramType.TOMO, False),
            (MammogramType.FFDM, MammogramType.SFM, True),
            (MammogramType.SYNTH, MammogramType.FFDM, False),
            (MammogramType.SYNTH, MammogramType.SYNTH, False),
            (MammogramType.SYNTH, MammogramType.TOMO, False),
            (MammogramType.TOMO, MammogramType.FFDM, True),
            (MammogramType.TOMO, MammogramType.SYNTH, True),
            (MammogramType.TOMO, MammogramType.TOMO, False),
            (MammogramType.TOMO, MammogramType.SFM, True),
            (MammogramType.UNKNOWN, MammogramType.UNKNOWN, False),
            (MammogramType.FFDM, MammogramType.UNKNOWN, True),
            (MammogramType.SYNTH, MammogramType.UNKNOWN, True),
            (MammogramType.TOMO, MammogramType.UNKNOWN, True),
            (MammogramType.SFM, MammogramType.UNKNOWN, True),
        ],
    )
    def test_lt(self, t1, t2, exp):
        assert (t1 < t2) == exp
        # TODO: Move these to separate tests if we decide to change the ordering
        # convention in the future.
        assert t1.is_preferred_to(t2) == exp
        assert MammogramType.get_best([t1, t2]) == (t1 if exp else t2)

    @pytest.mark.parametrize(
        "t1,t2,exp",
        [
            (MammogramType.FFDM, MammogramType.FFDM, True),
            (MammogramType.FFDM, MammogramType.SYNTH, True),
            (MammogramType.FFDM, MammogramType.TOMO, False),
            (MammogramType.FFDM, MammogramType.SFM, True),
            (MammogramType.SYNTH, MammogramType.FFDM, False),
            (MammogramType.SYNTH, MammogramType.SYNTH, True),
            (MammogramType.SYNTH, MammogramType.TOMO, False),
            (MammogramType.TOMO, MammogramType.FFDM, True),
            (MammogramType.TOMO, MammogramType.SYNTH, True),
            (MammogramType.TOMO, MammogramType.TOMO, True),
            (MammogramType.TOMO, MammogramType.SFM, True),
            (MammogramType.UNKNOWN, MammogramType.UNKNOWN, True),
            (MammogramType.FFDM, MammogramType.UNKNOWN, True),
            (MammogramType.SYNTH, MammogramType.UNKNOWN, True),
            (MammogramType.TOMO, MammogramType.UNKNOWN, True),
            (MammogramType.SFM, MammogramType.UNKNOWN, True),
        ],
    )
    def test_le(self, t1, t2, exp):
        assert (t1 <= t2) == exp

    @pytest.mark.parametrize(
        "t1,t2,exp",
        [
            (MammogramType.FFDM, MammogramType.FFDM, False),
            (MammogramType.FFDM, MammogramType.SYNTH, False),
            (MammogramType.FFDM, MammogramType.TOMO, True),
            (MammogramType.FFDM, MammogramType.SFM, False),
            (MammogramType.SYNTH, MammogramType.FFDM, True),
            (MammogramType.SYNTH, MammogramType.SYNTH, False),
            (MammogramType.SYNTH, MammogramType.TOMO, True),
            (MammogramType.TOMO, MammogramType.FFDM, False),
            (MammogramType.TOMO, MammogramType.SYNTH, False),
            (MammogramType.TOMO, MammogramType.TOMO, False),
            (MammogramType.TOMO, MammogramType.SFM, False),
            (MammogramType.UNKNOWN, MammogramType.UNKNOWN, False),
            (MammogramType.FFDM, MammogramType.UNKNOWN, False),
            (MammogramType.SYNTH, MammogramType.UNKNOWN, False),
            (MammogramType.TOMO, MammogramType.UNKNOWN, False),
            (MammogramType.SFM, MammogramType.UNKNOWN, False),
            (MammogramType.UNKNOWN, MammogramType.FFDM, True),
        ],
    )
    def test_gt(self, t1, t2, exp):
        assert (t1 > t2) == exp

    @pytest.mark.parametrize(
        "t1,t2,exp",
        [
            (MammogramType.FFDM, MammogramType.FFDM, True),
            (MammogramType.FFDM, MammogramType.SYNTH, False),
            (MammogramType.FFDM, MammogramType.TOMO, True),
            (MammogramType.FFDM, MammogramType.SFM, False),
            (MammogramType.SYNTH, MammogramType.FFDM, True),
            (MammogramType.SYNTH, MammogramType.SYNTH, True),
            (MammogramType.SYNTH, MammogramType.TOMO, True),
            (MammogramType.TOMO, MammogramType.FFDM, False),
            (MammogramType.TOMO, MammogramType.SYNTH, False),
            (MammogramType.TOMO, MammogramType.TOMO, True),
            (MammogramType.TOMO, MammogramType.SFM, False),
            (MammogramType.UNKNOWN, MammogramType.UNKNOWN, True),
            (MammogramType.FFDM, MammogramType.UNKNOWN, False),
            (MammogramType.SYNTH, MammogramType.UNKNOWN, False),
            (MammogramType.TOMO, MammogramType.UNKNOWN, False),
            (MammogramType.SFM, MammogramType.UNKNOWN, False),
            (MammogramType.UNKNOWN, MammogramType.FFDM, True),
        ],
    )
    def test_ge(self, t1, t2, exp):
        assert (t1 >= t2) == exp


class TestPhotometricInterpretation:
    @pytest.mark.parametrize(
        "val,expected",
        [
            pytest.param(PhotometricInterpretation.UNKNOWN, False),
            pytest.param(PhotometricInterpretation.MONOCHROME1, True),
            pytest.param(PhotometricInterpretation.MONOCHROME2, True),
            pytest.param(PhotometricInterpretation.RGB, True),
        ],
    )
    def test_bool(self, val, expected):
        assert bool(val) == expected

    @pytest.mark.parametrize(
        "val,expected",
        [
            pytest.param("", PhotometricInterpretation.UNKNOWN),
            pytest.param("MONOCHROME1", PhotometricInterpretation.MONOCHROME1),
            pytest.param("MONOCHROME2", PhotometricInterpretation.MONOCHROME2),
            pytest.param("RGB", PhotometricInterpretation.RGB),
        ],
    )
    def test_from_str(self, val, expected):
        pm = PhotometricInterpretation.from_str(val)
        assert pm == expected

    @pytest.mark.parametrize(
        "val,expected",
        [
            pytest.param(PhotometricInterpretation.UNKNOWN, False),
            pytest.param(PhotometricInterpretation.MONOCHROME1, True),
            pytest.param(PhotometricInterpretation.MONOCHROME2, True),
            pytest.param(PhotometricInterpretation.RGB, False),
        ],
    )
    def test_is_monochrome(self, val, expected):
        assert val.is_monochrome == expected

    @pytest.mark.parametrize(
        "val,expected",
        [
            pytest.param(None, PhotometricInterpretation.UNKNOWN),
            pytest.param("", PhotometricInterpretation.UNKNOWN),
            pytest.param("MONOCHROME1", PhotometricInterpretation.MONOCHROME1),
            pytest.param("MONOCHROME2", PhotometricInterpretation.MONOCHROME2),
            pytest.param("RGB", PhotometricInterpretation.RGB),
        ],
    )
    def test_from_dicom(self, dicom_object, val, expected):
        if val is not None:
            de = DataElement(Tag.PhotometricInterpretation, "CS", val)
            dicom_object[Tag.PhotometricInterpretation] = de
        else:
            del dicom_object[Tag.PhotometricInterpretation]

        pm = PhotometricInterpretation.from_dicom(dicom_object)
        assert pm == expected


class TestLaterality:
    @pytest.mark.parametrize(
        "orient,expected",
        [
            (Laterality.RIGHT, False),
            (Laterality.LEFT, False),
            (Laterality.BILATERAL, False),
            (Laterality.UNKNOWN, True),
            (Laterality.NONE, False),
        ],
    )
    def test_is_unknown(self, orient, expected):
        assert orient.is_unknown == expected

    @pytest.mark.parametrize(
        "orient,expected",
        [
            (Laterality.RIGHT, False),
            (Laterality.LEFT, False),
            (Laterality.BILATERAL, False),
            (Laterality.UNKNOWN, True),
            (Laterality.NONE, True),
        ],
    )
    def test_is_unknown_or_none(self, orient, expected):
        assert orient.is_unknown_or_none == expected

    @pytest.mark.parametrize(
        "orient,expected",
        [
            (Laterality.RIGHT, True),
            (Laterality.LEFT, True),
            (Laterality.BILATERAL, False),
            (Laterality.UNKNOWN, False),
            (Laterality.NONE, False),
        ],
    )
    def test_is_unilateral(self, orient, expected):
        assert orient.is_unilateral == expected

    @pytest.mark.parametrize(
        "l1,l2,expected",
        [
            (Laterality.RIGHT, Laterality.BILATERAL, Laterality.BILATERAL),
            (Laterality.LEFT, Laterality.RIGHT, Laterality.BILATERAL),
            (Laterality.RIGHT, Laterality.LEFT, Laterality.BILATERAL),
            (Laterality.LEFT, Laterality.LEFT, Laterality.LEFT),
            (Laterality.RIGHT, Laterality.NONE, Laterality.RIGHT),
            (Laterality.NONE, Laterality.NONE, Laterality.NONE),
            (Laterality.UNKNOWN, Laterality.NONE, Laterality.NONE),
            (Laterality.BILATERAL, Laterality.UNKNOWN, Laterality.BILATERAL),
        ],
    )
    def test_reduce(self, l1, l2, expected):
        assert l1.reduce(l2) == expected

    @pytest.mark.parametrize(
        "string,expected",
        [
            ("rcc", Laterality.RIGHT),
            ("lcc", Laterality.LEFT),
            ("rmlo", Laterality.RIGHT),
            ("lmlo", Laterality.LEFT),
            ("r-cc", Laterality.RIGHT),
            ("L-MLO", Laterality.LEFT),
            ("L CC", Laterality.LEFT),
            ("CCD", Laterality.RIGHT),
            ("CCE", Laterality.LEFT),
            ("MLOD", Laterality.RIGHT),
            ("MLOE", Laterality.LEFT),
            ("RML", Laterality.RIGHT),
            ("LML", Laterality.LEFT),
            ("foo", Laterality.UNKNOWN),
            ("bi", Laterality.BILATERAL),
            ("bilateral", Laterality.BILATERAL),
            ("", Laterality.UNKNOWN),
            ("none", Laterality.NONE),
        ],
    )
    def test_from_str(self, string, expected):
        orient = Laterality.from_str(string)
        assert orient == expected

    @pytest.mark.parametrize(
        "laterality,image_laterality,frame_laterality,expected",
        [
            ("L", None, None, Laterality.LEFT),
            (None, "L", None, Laterality.LEFT),
            (None, None, "L", Laterality.LEFT),
            ("L", "L", "L", Laterality.LEFT),
            ("R", None, None, Laterality.RIGHT),
            (None, "R", None, Laterality.RIGHT),
            (None, None, "R", Laterality.RIGHT),
            ("R", "R", "R", Laterality.RIGHT),
            (None, None, None, Laterality.UNKNOWN),
            ("", "", "", Laterality.UNKNOWN),
        ],
    )
    def test_from_tags(self, laterality, image_laterality, frame_laterality, expected):
        if frame_laterality:
            sfgs = [{Tag.FrameAnatomySequence: DummyElement([{Tag.FrameLaterality: DummyElement(frame_laterality)}])}]
        else:
            sfgs = None

        tags: Dict[int, Any] = {
            Tag.Laterality: laterality,
            Tag.ImageLaterality: image_laterality,
            Tag.SharedFunctionalGroupsSequence: sfgs,
        }
        tags = {k: v for k, v in tags.items() if v is not None}

        for k in tags:
            assert k in Laterality.get_required_tags()
        orient = Laterality.from_tags(tags)
        assert orient == expected

    @pytest.mark.parametrize(
        "value,expected",
        [
            (["A", "FL"], Laterality.RIGHT),
            (["A", "FR"], Laterality.LEFT),
            (["P", "FL"], Laterality.RIGHT),
            (["P", "FR"], Laterality.LEFT),
            (["A", "L"], Laterality.RIGHT),
            (["A", "R"], Laterality.LEFT),
            (["P", "L"], Laterality.RIGHT),
            (["P", "R"], Laterality.LEFT),
            (["P"], Laterality.UNKNOWN),
            ([], Laterality.UNKNOWN),
        ],
    )
    def test_from_patient_orientation(self, value, expected):
        orient = Laterality.from_patient_orientation(value)
        assert orient == expected

    def test_bool(self):
        expr = Laterality.UNKNOWN or Laterality.RIGHT or Laterality.UNKNOWN
        assert expr == Laterality.RIGHT

    def test_from_dicom(self, dicom_object):
        # trivial test since this wraps from_tags
        x = Laterality.from_dicom(dicom_object)
        assert x == Laterality.UNKNOWN

    @pytest.mark.parametrize(
        "lat,expected",
        [
            (Laterality.RIGHT, Laterality.LEFT),
            (Laterality.LEFT, Laterality.RIGHT),
            (Laterality.BILATERAL, Laterality.UNKNOWN),
            (Laterality.UNKNOWN, Laterality.UNKNOWN),
            (Laterality.NONE, Laterality.UNKNOWN),
        ],
    )
    def test_opposite(self, lat, expected):
        assert lat.opposite == expected


class TestViewPosition:
    @pytest.mark.parametrize(
        "orient,expected",
        [
            (ViewPosition.CC, False),
            (ViewPosition.CC, False),
            (ViewPosition.MLO, False),
            (ViewPosition.MLO, False),
            (ViewPosition.UNKNOWN, True),
        ],
    )
    def test_is_unknown(self, orient, expected):
        assert orient.is_unknown == expected

    @pytest.mark.parametrize(
        "string,expected",
        [
            ("rcc", ViewPosition.CC),
            ("lcc", ViewPosition.CC),
            ("rmlo", ViewPosition.MLO),
            ("lmlo", ViewPosition.MLO),
            ("r-cc", ViewPosition.CC),
            ("LMLO", ViewPosition.MLO),
            ("LCC", ViewPosition.CC),
            ("CCD", ViewPosition.CC),
            ("CCE", ViewPosition.CC),
            ("MLOD", ViewPosition.MLO),
            ("MLOE", ViewPosition.MLO),
            ("RML", ViewPosition.ML),
            ("LML", ViewPosition.ML),
            ("medio-lateral", ViewPosition.ML),
            ("medial-lateral", ViewPosition.ML),
            ("latero-medial", ViewPosition.LM),
            ("lateral-medial", ViewPosition.LM),
            ("cranio-caudal", ViewPosition.CC),
            ("caudal-cranial", ViewPosition.CC),
            ("medio-lateral oblique", ViewPosition.MLO),
            ("medial-lateral oblique", ViewPosition.MLO),
            ("latero-medial oblique", ViewPosition.LMO),
            ("lateral-medial oblique", ViewPosition.LMO),
            ("oblique medio-lateral", ViewPosition.MLO),
            ("oblique medial-lateral", ViewPosition.MLO),
            ("oblique latero-medial", ViewPosition.LMO),
            ("oblique lateral-medial", ViewPosition.LMO),
            ("cranio-caudal exaggerated laterally", ViewPosition.XCCL),
            ("cranio-caudal exaggerated medially", ViewPosition.XCCM),
            ("axillary tail", ViewPosition.AT),
            ("???", ViewPosition.UNKNOWN),
            ("foo", ViewPosition.UNKNOWN),
            ("", ViewPosition.UNKNOWN),
            ("rxccl", ViewPosition.XCCL),
            ("lxccl", ViewPosition.XCCL),
            ("rxccm", ViewPosition.XCCM),
            ("lxccm", ViewPosition.XCCM),
        ],
    )
    def test_from_str(self, string, expected):
        orient = ViewPosition.from_str(string)
        assert orient == expected

    @pytest.mark.parametrize(
        "val,expected",
        [
            ("MLO", ViewPosition.MLO),
            ("ML", ViewPosition.ML),
            ("CC", ViewPosition.CC),
            ("???", ViewPosition.UNKNOWN),
        ],
    )
    def test_from_tags(self, val, expected):
        tags = {0x00185101: val} if val is not None else {}
        for k in tags:
            assert k in ViewPosition.get_required_tags()
        orient = ViewPosition.from_tags(tags)
        assert orient == expected

    @pytest.mark.parametrize(
        "val,expected",
        [
            ("medio-lateral", ViewPosition.ML),
            ("medial-lateral", ViewPosition.ML),
            ("latero-medial", ViewPosition.LM),
            ("lateral-medial", ViewPosition.LM),
            ("cranio-caudal", ViewPosition.CC),
            ("caudal-cranial", ViewPosition.CC),
            ("medio-lateral oblique", ViewPosition.MLO),
            ("medial-lateral oblique", ViewPosition.MLO),
            ("latero-medial oblique", ViewPosition.LMO),
            ("lateral-medial oblique", ViewPosition.LMO),
            ("oblique medio-lateral", ViewPosition.MLO),
            ("oblique medial-lateral", ViewPosition.MLO),
            ("oblique latero-medial", ViewPosition.LMO),
            ("oblique lateral-medial", ViewPosition.LMO),
            ("cranio-caudal exaggerated laterally", ViewPosition.XCCL),
            ("cranio-caudal exaggerated medially", ViewPosition.XCCM),
            ("axillary tail", ViewPosition.AT),
            ("???", ViewPosition.UNKNOWN),
            ("", ViewPosition.UNKNOWN),
            (None, ViewPosition.UNKNOWN),
        ],
    )
    def test_from_view_code_sequence_tag(self, val, expected):
        view_code_sequence = [
            {
                "CodeMeaning": val,
            }
        ]
        assert Tag.ViewCodeSequence in ViewPosition.get_required_tags()
        orient = ViewPosition.from_view_code_sequence_tag(cast(DataElement, view_code_sequence))
        assert orient == expected

    @pytest.mark.parametrize(
        "value,expected",
        [
            (["A", "FL"], ViewPosition.MLO),
            (["A", "FR"], ViewPosition.MLO),
            (["P", "FL"], ViewPosition.MLO),
            (["P", "FR"], ViewPosition.MLO),
            (["A", "L"], ViewPosition.CC),
            (["A", "R"], ViewPosition.CC),
            (["P", "L"], ViewPosition.CC),
            (["P", "R"], ViewPosition.CC),
            (["P"], ViewPosition.UNKNOWN),
            ("CC", ViewPosition.CC),
            ("MLO", ViewPosition.MLO),
            ([], ViewPosition.UNKNOWN),
        ],
    )
    def test_from_patient_orientation(self, value, expected):
        orient = ViewPosition.from_patient_orientation(value)
        assert orient == expected

    def test_bool(self):
        expr = ViewPosition.UNKNOWN or ViewPosition.CC or ViewPosition.UNKNOWN
        assert expr == ViewPosition.CC

    @pytest.mark.parametrize(
        "view_pos,code,modifier_code,exp",
        [
            pytest.param("MLO", None, None, ViewPosition.MLO),
            pytest.param(None, "cranio-caudal", None, ViewPosition.CC),
            pytest.param(None, "medio-lateral oblique", "axillary tail", ViewPosition.AT),
        ],
    )
    def test_from_dicom(self, view_pos, code, modifier_code, exp):
        factory = DicomFactory(Modality="MG")
        overrides = {}
        if view_pos is not None:
            overrides["ViewPosition"] = view_pos
        if code is not None:
            overrides["ViewCodeSequence"] = DicomFactory.code_sequence(code)
        if modifier_code is not None:
            overrides["ViewModifierCodeSequence"] = DicomFactory.code_sequence(modifier_code)

        dcm = factory(**overrides)
        x = ViewPosition.from_dicom(dcm)
        assert x == exp

    @pytest.mark.parametrize(
        "value,expected",
        [
            (ViewPosition.MLO, True),
            (ViewPosition.CC, True),
            *[(view, False) for view in ViewPosition if view not in (ViewPosition.MLO, ViewPosition.CC)],
        ],
    )
    def test_is_standard_view(self, value, expected):
        assert value.is_standard_view == expected

    @pytest.mark.parametrize(
        "value,expected",
        [
            (view, view in (ViewPosition.MLO, ViewPosition.ML, ViewPosition.LM, ViewPosition.LMO))
            for view in ViewPosition
        ],
    )
    def test_is_mlo_like(self, value, expected):
        assert value.is_mlo_like == expected

    @pytest.mark.parametrize(
        "value,expected",
        [(view, view in (ViewPosition.CC, ViewPosition.XCCL, ViewPosition.XCCM)) for view in ViewPosition],
    )
    def test_is_cc_like(self, value, expected):
        assert value.is_cc_like == expected


class TestPixelSpacing:
    @pytest.mark.parametrize(
        "string,exp",
        [
            pytest.param("[0.01, 0.01]", PixelSpacing(0.01, 0.01)),
            pytest.param("[0.001, 0.01]", PixelSpacing(0.001, 0.01)),
            pytest.param("[1.0e-002, 2.0e-002]", PixelSpacing(0.01, 0.02)),
            pytest.param("[0.001-0.01]", PixelSpacing(0.001, 0.01)),
            pytest.param("[0.001;0.01]", PixelSpacing(0.001, 0.01)),
            pytest.param("[100, 100]", PixelSpacing(100, 100)),
            pytest.param("[001e002, 001e002]", PixelSpacing(100, 100)),
            pytest.param("", None, marks=pytest.mark.xfail(raises=ValueError)),
            pytest.param("[foo, bar]", None, marks=pytest.mark.xfail(raises=ValueError)),
        ],
    )
    def test_from_str(self, string, exp):
        actual = PixelSpacing.from_str(string)
        assert actual == exp

    @pytest.mark.parametrize(
        "tag,value,exp",
        [
            pytest.param(Tag.PixelSpacing, MultiValue(str, [0.01, 0.02]), PixelSpacing(0.01, 0.02)),
            pytest.param(Tag.ImagerPixelSpacing, MultiValue(str, [0.01, 0.02]), PixelSpacing(0.01, 0.02)),
        ],
    )
    def test_from_dicom_pixel_spacing(self, tag, value, exp):
        factory = DicomFactory(**{tag.name: value})
        dcm = factory()
        actual = PixelSpacing.from_dicom(dcm)
        assert actual == exp


class TestMammogramView:
    @pytest.mark.parametrize(
        "lat_in,view_in,lat_out,view_out",
        [
            (None, None, Laterality.UNKNOWN, ViewPosition.UNKNOWN),
            (Laterality.LEFT, None, Laterality.LEFT, ViewPosition.UNKNOWN),
            (None, ViewPosition.MLO, Laterality.UNKNOWN, ViewPosition.MLO),
        ],
    )
    def test_create(self, lat_in, view_in, lat_out, view_out):
        assert MammogramView.create(lat_in, view_in) == MammogramView(lat_out, view_out)

    @pytest.mark.parametrize(
        "lat_in,view_in,lat_out,view_out",
        [
            ("L", "MLO", Laterality.LEFT, ViewPosition.MLO),
            ("", "MLO", Laterality.UNKNOWN, ViewPosition.MLO),
            ("L", "", Laterality.LEFT, ViewPosition.UNKNOWN),
            ("R", "MLO", Laterality.RIGHT, ViewPosition.MLO),
        ],
    )
    def test_from_dicom(self, lat_in, view_in, lat_out, view_out):
        fact = DicomFactory(ViewPosition=view_in, Laterality=lat_in)
        dcm = fact()
        assert MammogramView.from_dicom(dcm) == MammogramView(lat_out, view_out)

    @pytest.mark.parametrize(
        "value,expected",
        [
            (view, view in (ViewPosition.MLO, ViewPosition.ML, ViewPosition.LM, ViewPosition.LMO))
            for view in ViewPosition
        ],
    )
    def test_is_mlo_like(self, value, expected):
        mview = MammogramView(view=value)
        assert mview.is_mlo_like == expected

    @pytest.mark.parametrize(
        "value,expected",
        [(view, view in (ViewPosition.CC, ViewPosition.XCCL, ViewPosition.XCCM)) for view in ViewPosition],
    )
    def test_is_cc_like(self, value, expected):
        mview = MammogramView(view=value)
        assert mview.is_cc_like == expected

    @pytest.mark.parametrize(
        "value,expected",
        [
            (ViewPosition.MLO, True),
            (ViewPosition.CC, True),
            *[(view, False) for view in ViewPosition if view not in (ViewPosition.MLO, ViewPosition.CC)],
        ],
    )
    def test_is_standard_view(self, value, expected):
        mview = MammogramView(view=value)
        assert mview.is_standard_mammo_view == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import importlib.metadata

# avoid BrokenPipeError, KeyboardInterrupt
import os
import warnings
from typing import Final, Iterable, Literal, cast

from .anonymize import anonymize
from .dicom import NoImageError, read_dicom_image
from .dicom_factory import DicomFactory
from .metadata import add_patient_age, dicom_to_json, drop_fields_by_length, get_date
from .volume import (
    VOLUME_HANDLERS,
    KeepVolume,
    RandomSlice,
    ReduceVolume,
    SliceAtLocation,
    UniformSample,
    VolumeHandler,
)


try:
    pass
except ImportError:
    print("DICOM operations require pydicom package")
    raise


__version__ = importlib.metadata.version("dicom-utils")


SPAM_WARNING_PATTERNS: Final = {
    ".*Invalid value for VR UI.*",
    ".*It's recommended that you change the Bits Stored value to produce the correct output.*",
}


def filter_spam_warnings(patterns: Iterable[str] = SPAM_WARNING_PATTERNS, action: str = "ignore") -> None:
    r"""Filters certain pydicom warnings that result in spam when manipulating DICOMS in bulk.
    If the environment variable ``SILENCE_PYDICOM_SPAM=1``, this function will be called
    automatically with ``action="ignore"`` when dicom-utils is imported.
    """
    for pattern in patterns:
        warnings.filterwarnings(cast(Literal["ignore"], action), pattern)


SILENCE_ENV_VAR: Final = "SILENCE_PYDICOM_SPAM"
if os.environ.get(SILENCE_ENV_VAR, "") == "1":
    filter_spam_warnings()


__all__ = [
    "__version__",
    "add_patient_age",
    "anonymize",
    "dicom_to_json",
    "DicomFactory",
    "drop_fields_by_length",
    "get_date",
    "filter_spam_warnings",
    "NoImageError",
    "read_dicom_image",
    "VolumeHandler",
    "KeepVolume",
    "SliceAtLocation",
    "UniformSample",
    "ReduceVolume",
    "RandomSlice",
    "VOLUME_HANDLERS",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import sys
from argparse import ArgumentParser
from pathlib import Path

from ..validation import Validator


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path", help="DICOM file to validate")
    parser.add_argument("-f", "--failing-only", default=False, action="store_true", help="only print failing tags")
    parser.add_argument("-n", "--no-color", default=False, action="store_true", help="disable printing with color")
    return parser


def main(args: argparse.Namespace) -> bool:
    src = Path(args.path)
    if not src.is_file():
        raise FileNotFoundError(src)

    validator = Validator.default_validator()
    results = validator.validate_dicom_file(src)
    validator.report_string(results, args.failing_only, not args.no_color)
    return validator.all_passing(results)


def entrypoint():
    result = main(get_parser().parse_args())
    sys.exit(0 if result else 1)


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Union

from pydicom.uid import UID


SeriesUID = UID
StudyUID = UID
SOPUID = UID
TransferSyntaxUID = UID
ImageUID = Union[SeriesUID, SOPUID]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import shutil
from pathlib import Path

import pytest

from dicom_utils.container.input import Input


@pytest.fixture
def suffix():
    return ".dcm"


@pytest.fixture
def dicom_files(tmp_path, dicom_file, suffix):
    paths = []
    for i in range(3):
        for j in range(3):
            dest = Path(tmp_path, f"subdir_{i}", f"file_{j}{suffix}")
            dest.parent.mkdir(exist_ok=True, parents=True)
            shutil.copy(dicom_file, str(dest))
            paths.append(dest)
    return paths


class TestInput:
    @pytest.mark.parametrize("suffix", [".dcm", "", ".123"])
    def test_basic_input(self, tmp_path, dicom_files):
        source = tmp_path
        Path(tmp_path, "dest")
        p = list(Input(source, use_bar=False))
        assert len(p) == 1
        key = p[0][0]
        assert isinstance(key, tuple)
        assert len(key) == 2
        assert all(isinstance(k, str) for k in key)



import shutil
from pathlib import Path

import pytest


@pytest.fixture
def dicom_folder(dicom_file, tmp_path):
    parent = Path(tmp_path)
    parent.mkdir(exist_ok=True)
    file_list = []
    for i in range(3):
        child = Path(dicom_file).with_suffix("").stem
        child = Path(parent, f"{child}_{i}").with_suffix(".dcm")
        shutil.copy(dicom_file, child)
        file_list.append(child)

    return file_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
from argparse import ArgumentParser
from pathlib import Path
from typing import Union

import numpy as np
import pydicom

from ..dicom import convert_frame_voi_lut, decompress


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path", type=Path, help="DICOM file to decompress")
    parser.add_argument("dest", type=Path, help="destination DICOM filepath")
    parser.add_argument("-m", "--method", default="max", choices=["max", "mean", "middle"], help="method of reduction")
    parser.add_argument(
        "-g", "--gpu", default=False, action="store_true", help="use NVJPEG2K accelerated decompression"
    )
    parser.add_argument(
        "-b", "--batch-size", default=4, type=int, help="batch size for NVJPEG2K accelerated decompression"
    )
    return parser


def safe_subtract(x: np.ndarray, other: Union[float, int, np.ndarray]) -> np.ndarray:
    return (x.astype(np.int64) - other).clip(min=0).astype(x.dtype)


def project(x: np.ndarray, reduction: str = "max") -> np.ndarray:
    VOLUME_DIM = -3
    if reduction == "max":
        x = x.max(axis=VOLUME_DIM)
    elif reduction == "mean":
        x = x.mean(axis=VOLUME_DIM).round().astype(x.dtype)
    elif reduction == "middle":
        L = x.shape[VOLUME_DIM]
        x = x[..., L // 2, :, :]
    return x


def main(args: argparse.Namespace) -> None:
    path = Path(args.path)
    dest = Path(args.dest)
    if not path.is_file():
        raise FileNotFoundError(path)
    if not dest.parent.is_dir():
        raise NotADirectoryError(dest.parent)

    with pydicom.dcmread(path) as dcm:
        dcm = decompress(dcm, strict=False, use_nvjpeg=args.gpu, batch_size=args.batch_size)
        dcm = convert_frame_voi_lut(dcm)
        pixels = project(dcm.pixel_array, args.method)
        dcm.PixelData = pixels.tobytes()
        dcm.NumberOfFrames = 1
        dcm.save_as(dest)


def entrypoint():
    parser = get_parser()
    main(parser.parse_args())


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import runpy
import shutil
import sys
from pathlib import Path

import pytest


@pytest.fixture
def dicom_file():
    pydicom = pytest.importorskip("pydicom")
    return pydicom.data.get_testdata_file("CT_small.dcm")


@pytest.fixture
def dicom_folder(dicom_file, tmp_path):
    parent = Path(tmp_path)
    parent.mkdir(exist_ok=True)
    file_list = []
    for i in range(3):
        child = Path(dicom_file).with_suffix("").stem
        child = Path(parent, f"{child}_{i}").with_suffix(".dcm")
        shutil.copy(dicom_file, child)
        file_list.append(child)

    return file_list


def test_find(dicom_folder, capsys, tmp_path):
    sys.argv = [
        sys.argv[0],
        str(tmp_path),
    ]
    runpy.run_module("dicom_utils.cli.find", run_name="__main__", alter_sys=True)
    captured1 = capsys.readouterr()
    runpy.run_module("dicom_utils.cli.find", run_name="__main__", alter_sys=True)
    captured2 = capsys.readouterr()

    for p in dicom_folder:
        assert str(p) in captured1.out
    assert list(sorted(captured1)) == list(sorted(captured2))


@pytest.mark.usefixtures("dicom_folder")
def test_find_pattern(capsys, tmp_path):
    sys.argv = [sys.argv[0], str(tmp_path), "--name", "*.DCM"]
    runpy.run_module("dicom_utils.cli.find", run_name="__main__", alter_sys=True)
    captured = capsys.readouterr()
    assert captured.out == ""


def test_find_parents(dicom_folder, capsys, tmp_path):
    sys.argv = [sys.argv[0], str(tmp_path), "--parents"]
    runpy.run_module("dicom_utils.cli.find", run_name="__main__", alter_sys=True)
    captured = capsys.readouterr()
    for p in dicom_folder:
        assert str(p.parent) in captured.out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import re
from typing import Optional, Protocol, Union, runtime_checkable

from .helpers import SOPUID, ImageUID, SeriesUID, StudyUID


logger = logging.getLogger(__name__)


@runtime_checkable
class SupportsStudyUID(Protocol):
    StudyInstanceUID: Optional[StudyUID] = None

    def same_study_as(self, other: "SupportsStudyUID") -> bool:
        r"""Checks if this record is part of the same study as record ``other``."""
        return bool(self.StudyInstanceUID) and self.StudyInstanceUID == other.StudyInstanceUID


@runtime_checkable
class SupportsPatientID(Protocol):
    PatientID: Optional[str] = None
    PatientName: Optional[str] = None

    def same_patient_as(self, other: "SupportsPatientID") -> bool:
        r"""Checks if this record references the same patient as record ``other``."""
        if self.PatientID or other.PatientID:
            return self.PatientID == other.PatientID
        else:
            return bool(self.PatientName) and self.PatientName == other.PatientName


@runtime_checkable
class SupportsStudyID(Protocol):
    StudyID: Optional[str] = None


@runtime_checkable
class SupportsUID(Protocol):
    SOPInstanceUID: Optional[SOPUID] = None
    SeriesInstanceUID: Optional[SeriesUID] = None

    @property
    def has_uid(self) -> bool:
        r"""Tests if the record has a SeriesInstanceUID or SOPInstanceUID"""
        return bool(self.SeriesInstanceUID or self.SOPInstanceUID)

    def same_uid_as(self, other: "SupportsUID", prefer_sop: bool = True) -> bool:
        r"""Checks if this record is part of the same study as record ``other``."""
        if not self.has_uid and other.has_uid:
            return False

        attr_to_check = "SOPInstanceUID" if prefer_sop else "SeriesInstanceUID"
        if (v_self := getattr(self, attr_to_check)) and (v_other := getattr(other, attr_to_check)):
            return v_self == v_other

        fallback_attr = "SOPInstanceUID" if not prefer_sop else "SeriesInstanceUID"
        assert attr_to_check != fallback_attr
        v_self = getattr(self, fallback_attr)
        v_other = getattr(other, fallback_attr)
        return bool(v_self) and bool(v_other) and v_self == v_other

    def get_uid(self, prefer_sop: bool = True) -> ImageUID:
        r"""Gets an image level UID. The UID will be chosen from SeriesInstanceUID and SOPInstanceUID,
        with preference as specified in ``prefer_sop``.
        """
        if not self.has_uid:
            raise AttributeError(f"{type(self)} has no UID")
        if prefer_sop:
            result = self.SOPInstanceUID or self.SeriesInstanceUID
        else:
            result = self.SeriesInstanceUID or self.SOPInstanceUID
        assert result is not None
        return result


@runtime_checkable
class SupportsStudyDate(Protocol):
    StudyDate: Optional[str] = None

    @property
    def StudyYear(self) -> Optional[int]:
        r"""Extracts a year from ``StudyDate``.

        Returns:
            First 4 digits of ``StudyDate`` as an int, or None if a year could not be parsed
        """
        if self.StudyDate and len(self.StudyDate) > 4:
            try:
                return int(self.StudyDate[:4])
            except Exception:
                pass
        return None


@runtime_checkable
class SupportsPatientAge(Protocol):
    PatientAge: Optional[str] = None
    PatientBirthDate: Optional[str] = None

    @property
    def patient_age(self) -> Optional[int]:
        r"""PatientAge as an int"""
        if not self.PatientAge:
            return None

        try:
            return int(re.sub(r"[^0-9]", "", self.PatientAge))
        except Exception:
            pass
        return None


@runtime_checkable
class SupportsGenerated(Protocol):
    generated: bool = False
    is_cad: bool = False


@runtime_checkable
class SupportsManufacturer(Protocol):
    Manufacturer: Optional[str] = None
    ManufacturerModelName: Optional[str] = None
    ManufacturerModelNumber: Optional[str] = None

    def search_manufacturer_info(self, pattern: Union[str, re.Pattern]) -> Optional[re.Match]:
        r"""Applies a regex search against known manufacturer info.

        Keys are searched in the following order:
            * ``Manufacturer``
            * ``ManufacturerModelName``
            * ``ManufacturerModelNumber``

        Args:
            pattern: A regex string or :class:`re.Pattern` to search with
        """
        pattern = pattern if isinstance(pattern, re.Pattern) else re.compile(pattern)
        assert isinstance(pattern, re.Pattern)
        if self.Manufacturer and (m := pattern.search(self.Manufacturer)):
            return m
        elif self.ManufacturerModelName and (m := pattern.search(self.ManufacturerModelName)):
            return m
        elif self.ManufacturerModelNumber and (m := pattern.search(self.ManufacturerModelNumber)):
            return m
        return None


@runtime_checkable
class SupportsSite(Protocol):
    InstitutionAddress: Optional[str] = None
    InstitutionName: Optional[str] = None
    TreatmentSite: Optional[str] = None

    @property
    def site(self) -> Optional[str]:
        return self.InstitutionAddress or self.TreatmentSite or self.InstitutionName


@runtime_checkable
class SupportsDataSetInfo(Protocol):
    DataSetSource: Optional[str] = None
    DataSetName: Optional[str] = None
    DataSetDescription: Optional[str] = None
    DataSetType: Optional[str] = None
    DataSetSubtype: Optional[str] = None


@runtime_checkable
class SupportsAccessionNumber(Protocol):
    AccessionNumber: Optional[str] = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
from pydicom.tag import Tag as PydicomTag

from dicom_utils.tags import Tag, create_tag, get_display_width


class TestTag:
    @pytest.mark.parametrize(
        "tag,expected",
        [
            pytest.param(Tag.SeriesInstanceUID, 0x0020000E),
            pytest.param(Tag.StudyInstanceUID, 0x0020000D),
        ],
    )
    def test_values(self, tag, expected):
        assert tag == expected

    def test_repr(self):
        t = Tag.PatientAge
        assert isinstance(str(t), str)
        assert isinstance(repr(t), str)

    @pytest.mark.parametrize(
        "tag,expected",
        [
            pytest.param(Tag.SeriesInstanceUID, 0x0020),
            pytest.param(Tag.StudyInstanceUID, 0x0020),
            pytest.param(Tag.EthnicGroup, 0x0010),
        ],
    )
    def test_group(self, tag, expected):
        assert tag.group == expected

    @pytest.mark.parametrize(
        "tag,expected",
        [
            pytest.param(Tag.SeriesInstanceUID, 0x000E),
            pytest.param(Tag.StudyInstanceUID, 0x000D),
        ],
    )
    def test_element(self, tag, expected):
        assert tag.element == expected

    @pytest.mark.parametrize(
        "tag,expected",
        [
            pytest.param(Tag.SeriesInstanceUID, (0x0020, 0x000E)),
            pytest.param(Tag.StudyInstanceUID, (0x0020, 0x000D)),
        ],
    )
    def test_tag_tuple(self, tag, expected):
        assert tag.tag_tuple == expected


@pytest.mark.parametrize(
    "val,tag",
    [
        pytest.param("StudyInstanceUID", Tag.StudyInstanceUID),
        pytest.param("SOPInstanceUID", Tag.SOPInstanceUID),
        pytest.param("EthnicGroup", Tag.EthnicGroup),
        pytest.param("PatientAge", Tag.PatientAge),
        pytest.param(Tag.PatientAge, Tag.PatientAge),
        pytest.param(PydicomTag("PatientAge"), Tag.PatientAge),
        pytest.param("FooBar", None, marks=pytest.mark.xfail(raises=ValueError)),
        pytest.param(1.0, None, marks=pytest.mark.xfail(raises=TypeError)),
    ],
)
def test_create_tag(val, tag):
    assert create_tag(val) == tag


def test_get_display_width():
    tags = [
        Tag.StudyInstanceUID,
        Tag.SeriesInstanceUID,
        Tag.PatientAge,
    ]
    expected = max(len(str(t)) for t in tags)
    assert get_display_width(tags) == expected
    assert get_display_width([]) == 0
